"""
A module for managing forms in the ArcGIS platform
"""

import json
from typing import Optional, Union
from arcgis import mapping
from arcgis._impl.common._mixins import PropertyMap
from arcgis.gis import Item
from arcgis.features import FeatureLayer
import copy


class FormCollection:
    """
    Represents a collection of forms in a webmap or item. A form is the editable counterpart to a popup
    -- it controls the appearance and behavior of your data collection experience in ArcGIS Field Maps
    and Map Viewer Beta.  These forms can then be used in the ArcGIS Field Maps mobile app and other applications.
    A form is stored as "formInfo" in the layer JSON on the webmap.
    This class will create a :class:`~arcgis.mapping.forms.FormInfo` object for each layer or table in the webmap/item data and return it
    as a list. You can then modify the :class:`~arcgis.mapping.forms.FormInfo` object to add and edit your form.

    ==================     ====================================================================
    **Parameter**           **Description**
    ------------------     --------------------------------------------------------------------
    parent                 Required :class:`~arcgis.mapping.WebMap` or :class:`~arcgis.gis.Item`.
                           This is the object which contains the layer, either an item of type
                           :class:`Feature Layer Collection <arcgis.features.FeatureLayerCollection>`
                           or a :class:`Web Map <arcgis.mapping.WebMap>`, where the forms are located.
                           This is needed to save your form changes to the backend.
    ==================     ====================================================================

    .. code-block:: python

        # USAGE EXAMPLE 1: Get Form Collection from WebMap
        wm = arcgis.mapping.WebMap(item)
        wm.add_layer(manhole_inspection)

        # get forms from webmap, get individual form from FormCollection, modify form
        form_collection = wm.forms
        form_info_2 = form_collection[0]
        form_info = form_collection.get(title="Manhole Inspection")
        form_info.clear()
        form_info.add_field(field_name="inspector", label="Inspector", description="This is the inspector")
        form_info.add_group(label="Group 1",initial_state="collapsed")
        form_info.update()

        # USAGE EXAMPLE 2: Create Form Collection

        from arcgis.mapping.forms import FormCollection
        wm = arcgis.mapping.WebMap(item)
        form_collection = FormCollection(wm)
        form_info_1 = form_collection.get(item_id="232323232323232323")
        form_info_1.title = "New Form"

    """

    def __init__(self, parent):
        self._parent = parent
        self._index = 0
        self._refresh_forms(parent)

    def __getitem__(self, key):
        return self.forms[key]

    def __iter__(self):
        return self

    def __next__(self):
        if self._index >= len(self.forms):
            raise StopIteration
        else:
            self._index += 1
            return self.forms[self._index - 1]

    def _refresh_forms(self, parent):
        if isinstance(parent, mapping.WebMap):
            self.forms = self._get_forms_from_webmap(parent)
        elif isinstance(parent, Item):
            if parent.type != "Feature Layer Collection":
                raise ValueError("Item must be feature layer collection to have forms")
            self.forms = self._get_forms_from_item(parent)
        else:
            raise ValueError("Parent item must be webmap or feature layer collection")

    def get(
        self,
        item_id: Optional[str] = None,
        title: Optional[str] = None,
        layer_id: Optional[str] = None,
    ):
        """
        Returns the form for the first layer with a matching item_id, title, or layer_id in the webmap's
        operational layers. Pass one of the three parameters into the method to return the form.

        ==================     ====================================================================
        **Parameter**           **Description**
        ------------------     --------------------------------------------------------------------
        item_id                Optional :class:`str`. Pass the item_id for the layer or table whose
                               form you'd like to return.
        ------------------     --------------------------------------------------------------------
        title                  Optional :class:`str`. Pass the title for the layer or table whose
                               form you'd like to return.
        ------------------     --------------------------------------------------------------------
        layer_id               Optional :class:`str`. Pass the layer_id for the layer whose form
                               you'd like to return
        ==================     ====================================================================

        :return: :class:`~arcgis.mapping.forms.FormInfo` or `None`
        """
        if item_id is None and title is None and layer_id is None:
            raise ValueError("Please pass at least one parameter into the function")
        self._refresh_forms(self._parent)
        if self.forms:
            for form in self.forms:
                # item id is optional in the webmap spec, so we need to try/except
                try:
                    if (
                        (title == form._layer_data["title"])
                        or (layer_id == form._layer_data["id"])
                        or (item_id == form._layer_data["itemId"])
                    ):
                        return form
                except Exception:
                    pass
        return None

    @staticmethod
    def _construct_forms_array(forms, layers, parent):
        """This is a shared method which creates an array of forms if given an array of layers."""
        # we save parent here into the form in order to distinguish between whether from_item or from_webmap was used and to get a usable GIS
        for layer in layers:
            FormCollection._get_forms_within_group_layer(forms, layer, parent)

    @staticmethod
    def _get_forms_within_group_layer(forms, layer, parent, subtype_gl_layer=None):
        if layer.layerType == "GroupLayer":
            for sub_layer in reversed(layer.layers):
                FormCollection._get_forms_within_group_layer(forms, sub_layer, parent)
        elif layer.layerType == "SubtypeGroupLayer":
            for sub_layer in reversed(layer.layers):
                FormCollection._get_forms_within_group_layer(
                    forms, sub_layer, parent, subtype_gl_layer=layer
                )
        elif layer.layerType == "ArcGISFeatureLayer" and not hasattr(
            layer, "featureCollection"
        ):
            forms.append(FormInfo(layer, parent, subtype_gl_layer))

    def _get_forms_from_item(self, item):
        """Populates self.forms given an item."""
        # we only support from_item for the instance where the form is saved into the feature layer item
        forms = []
        item_data = item.get_data()
        if "layers" in item_data:
            self._construct_forms_array(forms, item_data["layers"], parent=item)
        return forms

    def _get_forms_from_webmap(self, webmap):
        """Populates self.forms given a webmap."""
        forms = []
        self._construct_forms_array(forms, webmap.layers, parent=webmap)
        self._construct_forms_array(forms, webmap.tables, parent=webmap)
        return forms


class FormInfo:
    """
    Represents a form in ArcGIS Field Maps and other applications. This matches with
    the formInfo property in a webmap's operational layer.

    For more please see: https://developers.arcgis.com/web-map-specification/objects/formInfo/

    ==================     ====================================================================
    **Parameter**           **Description**
    ------------------     --------------------------------------------------------------------
    layer_data             Required :class:`PropertyMap` or :class:`dict`. This is the
                           operational layer which contains the formInfo dict. It can be
                           retrieved from a webmap using `arcgis.mapping.WebMap(item).layers[0]`
    ------------------     --------------------------------------------------------------------
    parent                 Required :class:`~arcgis.mapping.WebMap` or :class:`~arcgis.gis.Item`.
                           This is the object which contains the layer, either an item of type
                           `Feature Layer Collection` or a webmap. This is needed to save your
                           form changes to the backend.
    ------------------     --------------------------------------------------------------------
    subtype_gl_data        Optional :class:`PropertyMap` or :class:`dict`. This is the
                           operational layer representing a subtype group layer which contains
                           the layer containing the form. It can be retrieved from a webmap
                           using `arcgis.mapping.WebMap(item).layers[0]`
    ==================     ====================================================================

    .. code-block:: python

        # USAGE EXAMPLE 1: Modify FormInfo
            from arcgis.mapping.forms import FormFieldElement
            wm = arcgis.mapping.WebMap(item)
            wm.add_layer(manhole_inspection)
            form_collection = wm.forms
            form_info = form_collection.get_form(title="Manhole Inspection")

            # edit form properties
            form_info.title = "Manhole Inspection Form"
            form_info.description = "The editable experience in ArcGIS Field Maps for data collection"

            # get element, add element
            new_element = FormFieldElement(label="Inspector Name", field_name="inspectornm",
                                          editable=True)
            form_info.add(element=new_element)
            same_element = form_info.get(label="Inspector Name")

            # add group, add second group, add to group
            form_info.add_group(label="Inspector Group",description="This is a group for inspectors")
            new_group = FormGroupElement(label="Group 1", description="New Group")
            group = form_info.add(element=new_group)
            group.add_field(field_name="inspection_date", label="Inspection Date")

            # move element, delete element
            form_info.move(element=new_group, index=0)
            form_info.delete(element=new_group)

            # save form into backend
            form_info.update()
    """

    def __init__(self, layer_data, parent, subtype_gl_data=None, **kwargs):
        if not isinstance(layer_data, (dict, PropertyMap)):
            raise ValueError(
                "Incorrect layer type passed to FormInfo class. Please pass in a property map"
            )
        if subtype_gl_data and not isinstance(subtype_gl_data, (dict, PropertyMap)):
            raise ValueError(
                "Incorrect subtype group layer type passed to FormInfo class. Please pass in a property map"
            )
        self._kwargs = kwargs
        self._subtype_group_layer_data = subtype_gl_data
        self._original_layer = layer_data
        self._layer_data = copy.deepcopy(layer_data)
        self._form = self._layer_data.get("formInfo", {})
        self._title = self._form.get("title", self._layer_data.get("title", None))
        self._description = self._form.get("description")
        self._expression_infos = []
        for exp in self._form.get("expressionInfos", []):
            expression = FormExpressionInfo(
                expression=exp.get("expression"),
                name=exp.get("name"),
                title=exp.get("title"),
                return_type=exp.get("returnType"),
            )
            self._expression_infos.append(expression)
        self._form_elements = self._get_form_element_objects(
            self._form.get("formElements", []), self
        )
        self._parent = parent
        try:
            if self._subtype_group_layer_data:
                url = self._subtype_group_layer_data["url"]
            else:
                url = self._original_layer["url"]
            self.feature_layer = FeatureLayer(url=url, gis=self._parent._gis)
            self._fields = self._get_fields()
            self._edit_fields = self._get_edit_fields()
            self._id_fields = self._get_id_fields()
            self._required_fields = self._get_required_fields()
        except Exception:
            raise ValueError(
                "A layer url which can be used to generate a feature layer is required to use this module"
            )

    def __repr__(self):
        if self.title:
            return "Form: " + self.title
        else:
            return "Form"

    def __str__(self):
        if self.exists():
            return json.dumps(self.to_dict(), indent=2)
        return "None"

    def exists(self):
        """Returns whether the form exists for that particular layer or not."""
        return len(self._form_elements) > 0

    def clear(self):
        """Clears the form to an empty state. Deletes all form elements currently in the form."""
        self._expression_infos = []
        self._form_elements = []
        self._title = None
        self._description = None

    def get(self, label: Optional[str] = None):
        """
        Returns a matching FormElement given a label

        ==================     ====================================================================
        **Parameter**           **Description**
        ------------------     --------------------------------------------------------------------
        label                  Optional :class:`str`
                               The label of the form element you want to return from the list of
                               form elements
        ==================     ====================================================================

        :return: :class:`~arcgis.mapping.forms.FormFieldElement` or :class:`~arcgis.mapping.forms.FormGroupElement` or `None`
        """
        try:
            for el in self._form_elements:
                if el.label.lower() == label.lower():
                    return el
                if el.element_type == "group":
                    for grouped_field in el._form_elements:
                        if grouped_field.label.lower() == label.lower():
                            return grouped_field
        except Exception:
            return None

    def feature_layer(self):
        """The feature layer associated with the form"""
        return self.feature_layer

    @property
    def title(self):
        """Gets/sets the title of the form"""
        return self._title

    @title.setter
    def title(self, value):
        self._title = value

    @property
    def description(self):
        """Gets/sets the description of the form"""
        return self._description

    @description.setter
    def description(self, value):
        self._description = value

    @property
    def elements(self):
        """Returns elements in the form to the user - a list of :class:`~arcgis.mapping.forms.FormElement`"""
        return self._form_elements

    @property
    def expressions(self):
        """
        Returns Arcade expressions used in the form to the user - a list of :class:`~arcgis.mapping.forms.FormExpressionInfo`
        """
        return self._expression_infos

    def update(self):
        """
        Saves the form to the backend. If the form was derived from an :class:`~arcgis.gis.Item`, calling this function is required
        to save the form into the item. If the form was derived from a :class:`~arcgis.mapping.WebMap`, you can either call this
        function or :attr:`~arcgis.mapping.WebMap.update()`. If form has been cleared, removes formInfo from webmap
        """
        if self.exists():
            self._validate_all_required_fields_in_form()
        if isinstance(self._parent, Item):
            item_data = self._parent.get_data()
            if self.exists():
                item_data["layers"][self._layer_data["id"]]["formInfo"] = self.to_dict()
            else:
                item_data["layers"][self._layer_data["id"]].pop("formInfo", None)
            self._parent.update(data=item_data)
        if isinstance(self._parent, mapping.WebMap):
            if self.exists():
                self._original_layer["formInfo"] = self.to_dict()
            else:
                self._original_layer.pop("formInfo", None)
            try:
                self._parent.update()
            except RuntimeError:
                raise ValueError(
                    "WebMap item does not exist yet. Form is now on your webmap - please use WebMap.save() to persist these changes"
                )
        else:
            pass

    def to_dict(self):
        self._hydrate_expression_infos()
        data = {
            "expressionInfos": [exp.to_dict() for exp in self._expression_infos],
            "formElements": [element.to_dict() for element in self._form_elements],
        }
        if self._description:
            data["description"] = self._description
        if self._title:
            data["title"] = self._title
        for key, value in self._kwargs.items():
            data[key] = value
        return data

    def add_all_attributes(self):
        """Adds all fields which can be valid form elements (string, date, int, double, small) to the form."""
        fields = self._fields
        for field in fields:
            try:
                element = FormFieldElement(
                    label=field.get("alias"),
                    editable=field.get("editable"),
                    field_name=field.get("name"),
                )
                self.add(element=element)
            except Exception:
                continue

    def add(
        self,
        element=None,
        index: Optional[int] = None,
    ):
        """
        Adds a single :class:`~arcgis.mapping.forms.FormElement` to the form. You can add to the form either by instantiating your
        own :class:`~arcgis.mapping.forms.FormFieldElement` or :class:`~arcgis.mapping.forms.FormGroupElement` and passing it into the element parameter here,
        or you can type in a valid field_name in the form's layer and this function will attempt
        to add it to the form.

        ==================     ====================================================================
        **Parameter**           **Description**
        ------------------     --------------------------------------------------------------------
        element                Optional :class:`~arcgis.mapping.forms.FormFieldElement` or
                               :class:`~arcgis.mapping.forms.FormGroupElement`.
        ------------------     --------------------------------------------------------------------
        index                  Optional :class:`int`.
                               The index where you'd like the element in the form. If not provided,
                               this function will add the new element to the end of the form.
        ==================     ====================================================================

        :return: The element that was added - :class:`~arcgis.mapping.forms.FormFieldElement` or :class:`~arcgis.mapping.forms.FormGroupElement`
        """
        self._validate_input(element=element)
        self._validate_element(element)
        if index is None:
            index = len(self._form_elements)
        self._form_elements.insert(index, element)
        self._hydrate_expression_infos()
        return element

    def add_field(
        self,
        field_name: str,
        label: str,
        description: Optional[str] = None,
        visibility_expression=None,
        domain: Optional[dict] = None,
        editable: Optional[bool] = None,
        hint: Optional[str] = None,
        input_type: Optional[Union[str, dict]] = None,
        required_expression=None,
        index: Optional[int] = None,
        editable_expression=None,
        value_expression=None,
        **kwargs,
    ):
        """
        Adds a single field :class:`~arcgis.mapping.forms.FormElement` element to the end of the form.

        For more please see: https://developers.arcgis.com/web-map-specification/objects/formFieldElement/

        ======================     ====================================================================
        **Parameter**               **Description**
        ----------------------     --------------------------------------------------------------------
        field_name                 Required :class:`str`.
                                   The field name the form element corresponds to (where the data
                                   will be collected)
        ----------------------     --------------------------------------------------------------------
        label                      Required :class:`str`.
                                   The label of the form element
        ----------------------     --------------------------------------------------------------------
        description                Optional :class:`str`.
                                   The description of the form element
        ----------------------     --------------------------------------------------------------------
        visibility_expression      Optional :class:`~arcgis.mapping.forms.FormExpressionInfo`.
                                   The conditional visibility Arcade expression determining the
                                   visibility of the form element during data collection
        ----------------------     --------------------------------------------------------------------
        domain                     Optional :class:`dict`.
                                   The domain of the form element
        ----------------------     --------------------------------------------------------------------
        editable                   Optional :class:`bool`.
                                   Whether or not the form element is editable
        ----------------------     --------------------------------------------------------------------
        hint                       Optional :class:`str`.
                                   The hint for the user filling out the form element
        ----------------------     --------------------------------------------------------------------
        input_type                 Optional :class:`str` or :class:`dict`.
                                   The input type for the form element in ArcGIS Field Maps.
                                   Options include: "text-area", "text-box", "barcode-scanner",
                                   "combo-box", "radio-buttons", "datetime-picker"
        ----------------------     --------------------------------------------------------------------
        required_expression        Optional :class:`~arcgis.mapping.forms.FormExpressionInfo`.
                                   The conditional visibility Arcade expression determining the
                                   necessity of the form element during data collection
        ----------------------     --------------------------------------------------------------------
        index                      Optional :class:`int`.
                                   The index where you'd like the element in the form. If not provided,
                                   this function will add the new element to the end of the form.
        ----------------------     --------------------------------------------------------------------
        editable_expression        Optional :class:`~arcgis.mapping.forms.FormExpressionInfo`.
                                   The Arcade expression determining the
                                   editablity of the form element during data collection
        ----------------------     --------------------------------------------------------------------
        value_expression           Optional :class:`~arcgis.mapping.forms.FormExpressionInfo`.
                                   The Arcade expression which calculates a value for the form
                                   element during data collection
        ======================     ====================================================================

        :return: The element that was added - :class:`~arcgis.mapping.forms.FormGroupElement`
        """

        element = FormFieldElement(
            form=self,
            field_name=field_name,
            label=label,
            description=description,
            visibility_expression=visibility_expression,
            domain=domain,
            editable=editable,
            hint=hint,
            input_type=input_type,
            required_expression=required_expression,
            editable_expression=editable_expression,
            value_expression=value_expression,
            **kwargs,
        )
        return self.add(element, index=index)

    def add_group(
        self,
        label: str,
        description: Optional[str] = None,
        visibility_expression=None,
        initial_state: Optional[str] = None,
        index: Optional[int] = None,
        **kwargs,
    ):
        """
        Adds a single :class:`~arcgis.mapping.forms.GroupElement` to the form

        ======================     ====================================================================
        **Parameter**               **Description**
        ----------------------     --------------------------------------------------------------------
        label                      Required :class:`str`. The label of the group
        ----------------------     --------------------------------------------------------------------
        description                Optional :class:`str`. The description of the group
        ----------------------     --------------------------------------------------------------------
        visibility_expression      Optional :class:`~arcgis.mapping.forms.FormExpressionInfo`.
                                   The conditional visibility Arcade expression determining the
                                   visibility of the form element during data collection
        ----------------------     --------------------------------------------------------------------
        initial_state              Optional :class:`str`. The initial state of the group
        ----------------------     --------------------------------------------------------------------
        index                      Optional :class:`int`.
                                   The index where you'd like the element in the form. If not provided,
                                   this function will add the new element to the end of the form.
        ======================     ====================================================================

        :return: The element that was added - :class:`~arcgis.mapping.forms.FormGroupElement`
        """
        group_el = FormGroupElement(
            label=label,
            description=description,
            visibility_expression=visibility_expression,
            initial_state=initial_state,
            **kwargs,
        )
        return self.add(group_el, index=index)

    def delete(
        self,
        element=None,
        label: Optional[str] = None,
    ):
        """
        Deletes element from the form. You can use either the element param
        with a form element you get using :attr:`~arcgis.mapping.FormInfo.get()` or you can pass the label of the
        form element you'd like to move into the label param.

        ==================     ====================================================================
        **Parameter**           **Description**
        ------------------     --------------------------------------------------------------------
        element                Optional :class:`~arcgis.mapping.forms.FormFieldElement` or
                               :class:`~arcgis.mapping.forms.FormGroupElement`
        ------------------     --------------------------------------------------------------------
        label                  Optional :class:`str`.
                               An actual field name (not alias) corresponding to a field in the
                               form's feature layer
        ==================     ====================================================================

        :return: :class:`~arcgis.mapping.forms.FormFieldElement` or
                 :class:`~arcgis.mapping.forms.FormGroupElement`
                 or `False`
        """
        self._validate_input(element=element, field=label)
        if label:
            element = self.get(label=label)
        try:
            self._form_elements.remove(element)
            self._hydrate_expression_infos()
            return element
        except Exception:
            return False

    def move(
        self,
        element=None,
        label: Optional[str] = None,
        destination=None,
        index: Optional[int] = None,
    ):
        """
        Moves a form element in the form to a new location. You can use either the element param
        with a form element you get using :attr:`~arcgis.mapping.FormInfo.get()` or you can pass the label of the
        form element you'd like to move into the label param.

        ==================     ====================================================================
        **Parameter**           **Description**
        ------------------     --------------------------------------------------------------------
        element                Optional :class:`~arcgis.mapping.forms.FormFieldElement` or
                               :class:`~arcgis.mapping.forms.FormGroupElement`
        ------------------     --------------------------------------------------------------------
        label                  Optional :class:`str`
                               The label of the form element you'd like to move to a new location
        ------------------     ----------------------------------------------------------------------
        destination            Optional `arcgis.mapping.forms.FormGroupElement`
                               If moving an element into a group, please
                               provide this parameter. Otherwise, you will only be allowed to move
                               within the form.
        ------------------     --------------------------------------------------------------------
        index                  Optional :class:`int`.
                               The index where you'd like the element to move to in the form.
        ==================     ====================================================================

        :return: `True`
        """
        if index is None:
            raise ValueError("Please provide an index to this function")
        self._validate_input(element=element, field=label)
        if label:
            element = self.get(label=label)
        self.delete(element)
        if destination is None:
            destination = self
        return destination.add(element, index=index)

    def _get_matching_field(self, field_name):
        """Get the feature layer field given a popup field."""
        for field in self._fields:
            if field_name.lower() == field["name"].lower():
                return field
        else:
            raise ValueError("No matching field in the layer")

    def _get_fields(self):
        """Get feature layer's fields"""
        return self.feature_layer.properties.get("fields")

    @staticmethod
    def _is_valid_field_type(field_type):
        """Check the field type from the feature layer field is valid to be added to the form"""
        return field_type in [
            "esriFieldTypeDate",
            "esriFieldTypeDouble",
            "esriFieldTypeInteger",
            "esriFieldTypeSingle",
            "esriFieldTypeSmallInteger",
            "esriFieldTypeString",
            "esriFieldTypeGUID",
        ]

    @staticmethod
    def _get_default_input_type(field):
        """Gets the default input type based on the field type."""
        if field.get("domain", None):
            return "combo-box"
        if "date" in field["type"].lower():
            return "datetime-picker"
        return "text-box"

    @staticmethod
    def _validate_input(element=None, field=None):
        """Validate the inputs provided to add are valid."""
        if element and field:
            raise ValueError("Please use either element or field, not both")
        if element and not isinstance(element, FormElement):
            raise ValueError("Please pass a form element into the element parameter")
        if field and not isinstance(field, str):
            raise ValueError("Please pass a string into the element parameter")

    def _validate_element(self, element):
        """Validate the element passed to or created by add element is correct."""
        if element._form is None:
            element._form = self
        if not element.label:
            raise ValueError("Element must have label")
        if element.element_type == "field":
            matching_field = self._get_matching_field(element.field_name)
            if self._is_valid_field_type(matching_field["type"]):
                if not element.input_type:
                    element.input_type = self._get_default_input_type(matching_field)
            else:
                raise ValueError("Not a valid field type to add to the form")
            if self._is_geometry_field(element.field_name):
                raise ValueError("Cannot add a geometry field to the form")
            if not self._validate_unrestricted_field_name(element.field_name.lower()):
                raise ValueError(
                    "Cannot add a GPS metadata or editor tracking fields to the form"
                )
            for form_el in self._form_elements:
                if form_el.element_type == "group":
                    if element.field_name in [
                        el.field_name for el in form_el._form_elements
                    ]:
                        raise ValueError(
                            "Field already exists in a group, cannot add to form"
                        )
                else:
                    if element.field_name == form_el.field_name:
                        raise ValueError(
                            "Field already exists in the form, cannot add to the form"
                        )
        elif element.element_type == "group":
            for el in element.elements:
                self._validate_element(el)

    def _validate_unrestricted_field_name(self, field_name):
        """Validates the field is not a GPS metdata, edit, or id field."""
        return (
            "esrignss" not in field_name
            and "esrisnsr" not in field_name
            and field_name not in self._edit_fields
            and field_name not in self._id_fields
        )

    def _is_geometry_field(self, field_name):
        return "geometryProperties" in self.feature_layer.properties and (
            (
                "shapeAreaFieldName"
                in self.feature_layer.properties["geometryProperties"]
                and self.feature_layer.properties["geometryProperties"][
                    "shapeAreaFieldName"
                ].lower()
                == field_name.lower()
            )
            or (
                "shapeLengthFieldName"
                in self.feature_layer.properties["geometryProperties"]
                and self.feature_layer.properties["geometryProperties"][
                    "shapeLengthFieldName"
                ].lower()
                == field_name.lower()
            )
        )

    def _get_id_fields(self):
        """Returns the id fields in lower case."""
        try:
            return [
                self.feature_layer.properties.get("objectIdField").lower(),
                self.feature_layer.properties.get("globalIdField").lower(),
            ]
        except Exception:
            return []

    def _get_edit_fields(self):
        """Gets the edit fields for the feature layer in order to filter them out of the form."""
        try:
            return [
                x.lower()
                for x in list(
                    self.feature_layer.properties.get("editFieldsInfo", {}).values()
                )
            ]
        except Exception:
            return []

    @staticmethod
    def _get_expression_info(form, expression_name):
        return next((e for e in form.expressions if e.name == expression_name), None)

    @staticmethod
    def _get_form_element_objects(form_elements, form=None):
        """Shared between FormInfo and FormGroupElement to construct an array of FormElement objects from dictionaries for external usage."""
        elements = []
        for element in form_elements:
            if element["type"] == "field":
                el = FormFieldElement(
                    form=form,
                    description=element.get("description"),
                    label=element.get("label"),
                    visibility_expression=element.get("visibilityExpression"),
                    domain=element.get("domain"),
                    editable=element.get("editable"),
                    field_name=element.get("fieldName"),
                    hint=element.get("hint"),
                    input_type=element.get("inputType"),
                    required_expression=FormInfo._get_expression_info(
                        form, element.get("requiredExpression")
                    )
                    if form
                    else None,
                    editable_expression=FormInfo._get_expression_info(
                        form, element.get("editableExpression")
                    )
                    if form
                    else None,
                    value_expression=FormInfo._get_expression_info(
                        form, element.get("valueExpression")
                    )
                    if form
                    else None,
                )
            elif element["type"] == "group":
                el = FormGroupElement(
                    form=form,
                    elements=element.get("formElements"),
                    initial_state=element.get("initialState"),
                    description=element.get("description"),
                    label=element.get("label"),
                    visibility_expression=FormInfo._get_expression_info(
                        form, element.get("visibilityExpression")
                    )
                    if form
                    else None,
                )
            else:
                el = FormElement(
                    form=form,
                    element_type=element.get("type"),
                    description=element.get("description"),
                    label=element.get("label"),
                    visibility_expression=FormInfo._get_expression_info(
                        form, element.get("visibilityExpression")
                    )
                    if form
                    else None,
                )
            elements.append(el)
        return elements

    def _hydrate_expression_infos(self):
        for form_el in self._form_elements:
            self._hydrate_element_expressions(form_el)
            if form_el.element_type == "group":
                for el in form_el.elements:
                    self._hydrate_element_expressions(el)

    def _hydrate_element_expressions(self, element):
        """For each element's visibility and required expression, check if it exists in the expression info list. If it's a :class:`~arcgis.mapping.forms.FormExpressionInfo` object,
        add it to the list. If it's not, remove the expression as it does not point to anything and we can't form a :class:`~arcgis.mapping.forms.FormExpressionInfo`
        """
        if element.visibility_expression:
            if isinstance(element.visibility_expression, FormExpressionInfo):
                if (
                    self._get_expression_info(self, element.visibility_expression.name)
                    is None
                ):
                    self._expression_infos.append(element.visibility_expression)
            else:
                if (
                    self._get_expression_info(self, element.visibility_expression)
                    is None
                ):
                    element._visibility_expression = None
        if element.element_type == "field":
            if element.required_expression:
                if isinstance(element.required_expression, FormExpressionInfo):
                    if (
                        self._get_expression_info(
                            self, element.required_expression.name
                        )
                        is None
                    ):
                        self._expression_infos.append(element.required_expression)
                else:
                    if (
                        self._get_expression_info(self, element.required_expression)
                        is None
                    ):
                        element._required_expression = None
            if element.editable_expression:
                if isinstance(element.editable_expression, FormExpressionInfo):
                    if (
                        self._get_expression_info(
                            self, element.editable_expression.name
                        )
                        is None
                    ):
                        self._expression_infos.append(element.editable_expression)
                else:
                    if (
                        self._get_expression_info(self, element.editable_expression)
                        is None
                    ):
                        element._editable_expression = None
            if element.value_expression:
                if isinstance(element.value_expression, FormExpressionInfo):
                    if (
                        self._get_expression_info(self, element.value_expression.name)
                        is None
                    ):
                        self._expression_infos.append(element.value_expression)
                else:
                    if (
                        self._get_expression_info(self, element.value_expression)
                        is None
                    ):
                        element._value_expression = None

    def _get_required_fields(self):
        required_fields = []
        for field in self._fields:
            if (
                "nullable" in field
                and field["nullable"] is False
                and field["name"].lower() not in self._edit_fields
                and field["name"].lower() not in self._id_fields
            ):
                required_fields.append(field["name"])
        return required_fields

    def _validate_all_required_fields_in_form(self):
        found = False
        for field_name in self._required_fields:
            for el in self._form_elements:
                if el.element_type == "group":
                    if field_name.lower() in [
                        el.field_name.lower() for el in el._form_elements
                    ]:
                        found = True
                        break
                else:
                    if field_name.lower() == el.field_name.lower():
                        found = True
                        break
            if not found:
                raise ValueError(
                    str(field_name)
                    + " is a required field not found in the form. Please add to the form"
                )
            found = False


class FormElement:
    """
    The superclass class for :class:`~arcgis.mapping.forms.FormFieldElement` and :class:`~arcgis.mapping.forms.FormGroupElement`. Contains properties common to
    the two types of field elements. Instantiate a FormFieldElement or FormGroupElement instead of this class.
    """

    def __init__(
        self,
        form=None,
        element_type=None,
        description=None,
        label=None,
        visibility_expression=None,
        **kwargs,
    ):
        self._form = form
        self._element_type = element_type
        self._description = description
        self._label = label
        self._visibility_expression = visibility_expression
        self._kwargs = kwargs

    @property
    def description(self):
        """Gets/sets the description of the form element."""
        return self._description

    @description.setter
    def description(self, value):
        self._description = value

    @property
    def label(self):
        """Gets/sets the label of the form element."""
        return self._label

    @label.setter
    def label(self, value):
        self._label = value

    @property
    def element_type(self):
        """Get/Set the element type of the form element."""
        return self._element_type

    @element_type.setter
    def element_type(self, value):
        if self.element_type is None:
            self._element_type = value
        else:
            raise ValueError("Cannot change element type once it has been set")

    @property
    def visibility_expression(self):
        """Gets/sets the visibility expression of the form element."""
        return self._visibility_expression

    @visibility_expression.setter
    def visibility_expression(self, value):
        if isinstance(value, FormExpressionInfo) or value is None:
            self._visibility_expression = value
        else:
            raise ValueError("Please pass a FormExpressionInfo object")

    def to_dict(self):
        el_dict = {}
        if self._description:
            el_dict["description"] = self._description
        if self._label:
            el_dict["label"] = self._label
        if self._element_type:
            el_dict["type"] = self._element_type
        if self._visibility_expression:
            el_dict["visibilityExpression"] = self._visibility_expression.name
        for key, value in self._kwargs.items():
            el_dict[key] = value
        return el_dict


class FormFieldElement(FormElement):
    """
    Represents a single field (non-group) element in a form. This corresponds with a field
    in the feature layer where you are collecting data. This is a subclass of FormElement, so
    you can modify properties such as label, description, and visibility_expression on these
    objects as well.

    For more please see: https://developers.arcgis.com/web-map-specification/objects/formFieldElement/

    ======================     ====================================================================
    **Parameter**               **Description**
    ----------------------     --------------------------------------------------------------------
    form                       Optional :class:`~arcgis.mapping.forms.FormInfo`.
                               The form which contains this field element.
    ----------------------     --------------------------------------------------------------------

    description                Optional :class:`str`.
                               The description of the form element
    ----------------------     --------------------------------------------------------------------
    label                      Optional :class:`str`.
                               The label of the form element
    ----------------------     --------------------------------------------------------------------
    visibility_expression      Optional :class:`~arcgis.mapping.forms.FormExpressionInfo`.
                               The conditional visibility Arcade expression determining the
                               visibility of the form element during data collection
    ----------------------     --------------------------------------------------------------------
    domain                     Optional :class:`dict`.
                               The domain of the form element
    ----------------------     --------------------------------------------------------------------
    editable                   Optional :class:`bool`.
                               Whether or not the form element is editable
    ----------------------     --------------------------------------------------------------------
    field_name                 Optional :class:`str`.
                               The field name the form element corresponds to (where the data
                               will be collected)
    ----------------------     --------------------------------------------------------------------
    hint                       Optional :class:`str`.
                               The hint for the user filling out the form element
    ----------------------     --------------------------------------------------------------------
    input_type                 Optional :class:`str` or :class:`dict`.
                               The input type for the form element in ArcGIS Field Maps.

                               Options include:
                                    "text-area", "text-box", "barcode-scanner", "combo-box", "radio-buttons", "datetime-picker"
    ----------------------     --------------------------------------------------------------------
    required_expression        Optional :class:`~arcgis.mapping.forms.FormExpressionInfo`.
                               The conditional visibility Arcade expression determining the
                               necessity of the form element during data collection
    ======================     ====================================================================

    .. code-block:: python

        # USAGE EXAMPLE 1: Edit properties on form element
        from arcgis.mapping.forms import FormExpressionInfo
        wm = arcgis.mapping.WebMap(item)
        wm.add_layer(manhole_inspection)
        form_collection = wm.forms
        form_info = form_collection.get_form(title="Manhole Inspection")

        # edit element properties
        form_element = form_info.get(label="Inspector Name")
        form_element.label = "Inspector Name(s)"
        form_element.description = "The inspector(s) who completed this manhole inspection")

        # set visibility expression
        el = form_info.add_field(field_name="jake_only", label="jake_only")
        expression_info = FormExpressionInfo(name="expr0",title="New Expression",expression="$feature.inspector == 'Jake'")
        el.visibility_expression = expression_info
    """

    def __init__(
        self,
        form=None,
        description=None,
        label=None,
        visibility_expression=None,
        domain=None,
        editable=None,
        field_name=None,
        hint=None,
        input_type=None,
        required_expression=None,
        editable_expression=None,
        value_expression=None,
        **kwargs,
    ):
        super().__init__(
            form=form,
            element_type="field",
            description=description,
            label=label,
            visibility_expression=visibility_expression,
            **kwargs,
        )
        self._domain = domain
        self._editable = editable
        self._field_name = field_name
        self._hint = hint
        self.input_type = input_type
        self._required_expression = required_expression
        self._editable_expression = editable_expression
        self._value_expression = value_expression

    def __repr__(self):
        if self._label:
            return "Field " + self._label
        else:
            return "Field"

    def __str__(self):
        return json.dumps(self.to_dict(), indent=2)

    @property
    def domain(self):
        """Gets/sets the domain of the form element."""
        return self._domain

    @domain.setter
    def domain(self, value: dict):
        if not isinstance(value, dict):
            raise ValueError("Please pass a dict into this function")
        self._domain = value

    @property
    def editable(self):
        """Gets/sets the editability of the form element."""
        return self._editable

    @editable.setter
    def editable(self, value: bool):
        self._editable = value

    @property
    def field_name(self):
        """Gets the field name for the form element."""
        if self._field_name:
            return self._field_name.lower()
        else:
            return None

    @field_name.setter
    def field_name(self, value):
        if self._field_name is None:
            self._field_name = value
        else:
            raise ValueError("Cannot modify field name once it has been set")

    @property
    def hint(self):
        """Gets/sets the hint of the form element."""
        return self._hint

    @hint.setter
    def hint(self, value):
        self._hint = value

    @property
    def input_type(self):
        """
        Gets/sets the input type of the form element.

        ===============     ====================================================================
        **Parameter**        **Description**
        ---------------     --------------------------------------------------------------------
        value               Required string or dictionary.

                            Values:

                                "text-area" | "text-box" | "barcode-scanner" | "combo-box" | "radio-buttons" | "datetime-picker"
        ===============     ====================================================================

        :return: dictionary that represents the input type
        """
        return self._input_type

    @input_type.setter
    def input_type(self, value: Union[str, dict]):
        if value in [
            "text-area",
            "text-box",
            "barcode-scanner",
            "combo-box",
            "radio-buttons",
            "datetime-picker",
        ]:
            value = {"type": value}
        self._input_type = value

    @property
    def required_expression(self):
        """Gets/sets the required expression of the form element. Takes an object of :class:`~arcgis.mapping.forms.FormExpressionInfo` ."""
        return self._required_expression

    @required_expression.setter
    def required_expression(self, value):
        if isinstance(value, FormExpressionInfo) or value is None:
            self._required_expression = value
        else:
            raise ValueError("Please pass a FormExpressionInfo object")

    @property
    def editable_expression(self):
        """Gets/sets the editable expression of the form element. Takes an object of :class:`~arcgis.mapping.forms.FormExpressionInfo` ."""
        return self._editable_expression

    @editable_expression.setter
    def editable_expression(self, value):
        if isinstance(value, FormExpressionInfo) or value is None:
            self._editable_expression = value
        else:
            raise ValueError("Please pass a FormExpressionInfo object")

    @property
    def value_expression(self):
        """Gets/sets the value expression of the form element. Takes an object of :class:`~arcgis.mapping.forms.FormExpressionInfo`  ."""
        return self._value_expression

    @value_expression.setter
    def value_expression(self, value):
        if isinstance(value, FormExpressionInfo) or value is None:
            self._value_expression = value
        else:
            raise ValueError("Please pass a FormExpressionInfo object")

    def to_dict(self):
        el_dict = super().to_dict()
        if self._domain:
            el_dict["domain"] = self._domain
        if self._editable:
            el_dict["editable"] = self._editable
        if self._field_name:
            el_dict["fieldName"] = self._field_name
        if self._hint:
            el_dict["hint"] = self._hint
        if self._input_type:
            el_dict["inputType"] = self._input_type
        if self._required_expression:
            el_dict["requiredExpression"] = self._required_expression.name
        if self._editable_expression:
            el_dict["editableExpression"] = self._editable_expression.name
        if self._value_expression:
            el_dict["valueExpression"] = self._value_expression.name
        return el_dict


class FormGroupElement(FormElement):
    """
    Represents a single group element in a form. This is a subclass of FormElement, so
    you can modify properties such as label, description, and visibility_expression on these
    objects as well.

    For more please see: https://developers.arcgis.com/web-map-specification/objects/formGroupElement/

    ======================     ====================================================================
    **Parameter**               **Description**
    ----------------------     --------------------------------------------------------------------
    form                       Optional :class:`~arcgis.mapping.forms.FormInfo`.
                               The form which contains this group element.
    ----------------------     --------------------------------------------------------------------
    description                Optional :class:`str`.
                               The description of the group element
    ----------------------     --------------------------------------------------------------------
    label                      Optional :class:`str`.
                               The label of the group element
    ----------------------     --------------------------------------------------------------------
    visibility_expression      Optional :class:`~arcgis.mapping.forms.FormExpressionInfo`.
                               The conditional visibility Arcade expression determining the
                               visibility of the form element during data collection
    ----------------------     --------------------------------------------------------------------
    initial_state              Optional :class:`dict`.
                               Options are "collapsed" and "expanded"
    ======================     ====================================================================

    .. code-block:: python

        # USAGE EXAMPLE 1: Edit properties on group, add to group
        from arcgis.mapping.forms import FormExpressionInfo
        wm = arcgis.mapping.WebMap(item)
        wm.add_layer(manhole_inspection)
        form_collection = wm.forms
        form_info = form_collection.get_form(title="Manhole Inspection")

        # edit group properties, access elements within group
        group_element = form_info.get(label="Group 1")
        grouped_form_element = group_element.get(label="Inspector Name")
        grouped_form_element.label = "Inspector Name(s)
        group_element.label = "Inspector Information"
        group_element.initial_state = "collapsed"

        # add group, add to group, delete from group, delete group
        new_group = FormGroupElement(form_info, label="Group 2", initial_state="expanded")
        group = form_info.add(element=new_group)
        grouped_element = group.add_field(field_name="inspection_date", label="Inspection Date")
        group.add_field(field_name="inspection_city", label="Inspection City")
        grouped_element.label = "Inspection Date"
        group.move(grouped_element, index=1)
        group.delete(grouped_element)
        form_info.delete(group)

    """

    def __init__(
        self,
        form=None,
        elements=None,
        initial_state=None,
        description=None,
        label=None,
        visibility_expression=None,
        **kwargs,
    ):
        super().__init__(
            form=form,
            element_type="group",
            description=description,
            label=label,
            visibility_expression=visibility_expression,
            **kwargs,
        )
        if elements is None:
            elements = []
        self._form_elements = FormInfo._get_form_element_objects(elements, form=form)
        self._initial_state = initial_state

    def __repr__(self):
        return "Group: " + self._label

    def __str__(self):
        return json.dumps(self.to_dict(), indent=2)

    @property
    def elements(self):
        """Returns elements in the group to the user - a list of :class:`~arcgis.mapping.forms.FormElement`."""
        return self._form_elements

    @property
    def initial_state(self):
        """
        Gets/sets the initial state of the form element.

        ===============     ====================================================================
        **Parameter**        **Description**
        ---------------     --------------------------------------------------------------------
        value               Required string.

                            Values:

                                "collapsed" | "expanded"
        ===============     ====================================================================
        """
        return self._initial_state

    @initial_state.setter
    def initial_state(self, value: str):
        if value not in ["collapsed", "expanded"]:
            raise ValueError("Value can either be collapsed or expanded")
        self._initial_state = value

    def add(
        self, element: Optional[FormFieldElement] = None, index: Optional[int] = None
    ):
        """
        Adds a single form element to the group. You can add to the group either by instantiating
        a FormFieldElement and passing it into the element parameter here,
        or you can type in a valid field_name in the form's layer and this function will attempt
        to add it to the form.

        ==================     ====================================================================
        **Parameter**           **Description**
        ------------------     --------------------------------------------------------------------
        element                Optional :class:`~arcgis.mapping.forms.FormFieldElement`
        ------------------     --------------------------------------------------------------------
        index                  Optional :class:`int`.
                               The index where you'd like the element in the group. If not provided,
                               this function will add the new element to the end of the group.
        ==================     ====================================================================

        :return: The element that was added - :class:`~arcgis.mapping.forms.FormFieldElement`
        """
        self._validate_element(element)
        if index is None:
            index = len(self._form_elements)
        self._form_elements.insert(index, element)
        self._form._hydrate_expression_infos()
        return element

    def add_field(
        self,
        field_name: str,
        label: str,
        description: Optional[str] = None,
        visibility_expression=None,
        domain: Optional[dict] = None,
        editable: Optional[bool] = None,
        hint: Optional[str] = None,
        input_type: Optional[Union[dict, str]] = None,
        required_expression=None,
        index: Optional[int] = None,
        editable_expression=None,
        value_expression=None,
        **kwargs,
    ):
        """
        Adds a single field :class:`~arcgis.mapping.forms.FormElement` element to the end of the group.

        For more please see: https://developers.arcgis.com/web-map-specification/objects/formFieldElement/

        ======================     ====================================================================
        **Parameter**               **Description**
        ----------------------     --------------------------------------------------------------------
        field_name                 Required :class:`str`.
                                   The field name the form element corresponds to (where the data
                                   will be collected)
        ----------------------     --------------------------------------------------------------------
        label                      Required :class:`str`.
                                   The label of the form element
        ----------------------     --------------------------------------------------------------------
        description                Optional :class:`str`.
                                   The description of the form element
        ----------------------     --------------------------------------------------------------------
        visibility_expression      Optional :class:`~arcgis.mapping.forms.FormExpressionInfo`.
                                   The conditional visibility Arcade expression determining the
                                   visibility of the form element during data collection
        ----------------------     --------------------------------------------------------------------
        domain                     Optional :class:`dict`.
                                   The domain of the form element
        ----------------------     --------------------------------------------------------------------
        editable                   Optional :class:`bool`.
                                   Whether or not the form element is editable
        ----------------------     --------------------------------------------------------------------
        hint                       Optional :class:`str`.
                                   The hint for the user filling out the form element
        ----------------------     --------------------------------------------------------------------
        input_type                 Optional :class:`str` or :class:`dict`.
                                   The input type for the form element in ArcGIS Field Maps.
                                   Options include: "text-area", "text-box", "barcode-scanner",
                                   "combo-box", "radio-buttons", "datetime-picker"
        ----------------------     --------------------------------------------------------------------
        required_expression        Optional :class:`~arcgis.mapping.forms.FormExpressionInfo`.
                                   The conditional visibility Arcade expression determining the
                                   necessity of the form element during data collection
        ----------------------     --------------------------------------------------------------------
        index                      Optional :class:`int`.
                                   The index where you'd like the element in the form. If not provided,
                                   this function will add the new element to the end of the form.
        ----------------------     --------------------------------------------------------------------
        editable_expression        Optional :class:`~arcgis.mapping.forms.FormExpressionInfo`.
                                   The Arcade expression determining how much editable the form element is during data collection
        ----------------------     --------------------------------------------------------------------
        value_expression           Optional :class:`~arcgis.mapping.forms.FormExpressionInfo`.
                                   The Arcade expression which calculates a value for the form
                                   element during data collection
        ======================     ====================================================================

        """
        element = FormFieldElement(
            form=self,
            field_name=field_name,
            label=label,
            description=description,
            visibility_expression=visibility_expression,
            domain=domain,
            editable=editable,
            hint=hint,
            input_type=input_type,
            required_expression=required_expression,
            editable_expression=editable_expression,
            value_expression=value_expression,
            **kwargs,
        )
        return self.add(element, index=index)

    def delete(
        self, element: Optional[FormFieldElement] = None, label: Optional[str] = None
    ):
        """
        Deletes form element from the group. You can use either the element param
        with a form element you get using `FormInfo.get()` or you can pass the label of the
        form element you'd like to move into the label param.

        ==================     ====================================================================
        **Parameter**           **Description**
        ------------------     --------------------------------------------------------------------
        element                Optional :class:`~arcgis.mapping.forms.FormFieldElement`
        ------------------     --------------------------------------------------------------------
        label                  Optional :class:`str`
                               An actual field name (not alias) corresponding to a field in the
                               form's feature layer
        ==================     ====================================================================

        :return: The deleted element - :class:`~arcgis.mapping.forms.FormFieldElement` or `False`
        """
        if label:
            element = self.get(label=label)
        try:
            self._form_elements.remove(element)
            return element
        except Exception:
            return False

    def move(
        self,
        element: Optional[FormFieldElement] = None,
        label: Optional[str] = None,
        destination=None,
        index: Optional[int] = None,
    ):
        """
        Moves a form element in the group to a new location. You can use either the element param
        with a form element you get using `FormGroupElement.get() <:attr:`arcgis.mapping.forms.FormGroupElement.get`>` or you can pass the label
        of the form element you'd like to move into the label param.

        ==================     ====================================================================
        **Parameter**           **Description**
        ------------------     --------------------------------------------------------------------
        element                Optional :class:`~arcgis.mapping.forms.FormFieldElement`
        ------------------     --------------------------------------------------------------------
        label                  Optional :class:`str`
                               The label of the form element you'd like to move to a new location
        ------------------     ----------------------------------------------------------------------
        destination            Optional :class:`~arcgis.mapping.forms.FormInfo` or
                               `arcgis.mapping.forms.FormGroupElement`
                               If moving out of the group (to the form or to new group), please
                               provide this parameter. Otherwise, you will only be allowed to move
                               within this group.
        ------------------     --------------------------------------------------------------------
        index                  Optional :class:`int`.
                               The index where you'd like the element to move to in the group.
        ==================     ====================================================================

        :return: The element that was moved - :class:`~arcgis.mapping.forms.FormFieldElement`
        """
        if index is None:
            raise ValueError("Please provide an index")
        if label:
            element = self.get(label=label)
        self.delete(element)
        if destination is None:
            destination = self
        return destination.add(element, index=index)

    def get(self, label: Optional[str] = None):
        """
        Returns a matching FormFieldElement in the group given a label

        ==================     ====================================================================
        **Parameter**           **Description**
        ------------------     --------------------------------------------------------------------
        label                  Optional :class:`str`
                               The label of the form element you want to return from the list of
                               form elements
        ==================     ====================================================================

        :return: :class:`~arcgis.mapping.forms.FormFieldElement` or `None`
        """
        try:
            return next(
                el for el in self._form_elements if el.label.lower() == label.lower()
            )
        except Exception:
            return None

    def to_dict(self):
        el_dict = super().to_dict()
        # don't use if form_elements here as the empty array is required
        el_dict["formElements"] = [element.to_dict() for element in self._form_elements]
        if self._initial_state:
            el_dict["initialState"] = self._initial_state
        return el_dict

    def _validate_element(self, element):
        """Validate the element passed to or created by add element is correct"""
        if element.element_type == "group":
            raise ValueError("You cannot add a group to another group")
        if self._form is not None:
            self._form._validate_element(element)


class FormExpressionInfo:
    """
    This class corresponds to a single expressionInfo in the expressionInfos list within a form.

    For more please see: https://developers.arcgis.com/web-map-specification/objects/formExpressionInfo/

    ==================     ====================================================================
    **Parameter**           **Description**
    ------------------     --------------------------------------------------------------------
    expression             Optional :class:`str`
                           This is an Arcade expression which you want to evaluate in ArcGIS
                           Field Maps on whether or not this form element shows.
    ------------------     --------------------------------------------------------------------
    name                   Optional :class:`str`
                           The unique identifier for the expressionInfo
    ------------------     --------------------------------------------------------------------
    title                  Optional :class:`int`.
                           The user friendly name for the expressionInfo
    ------------------     --------------------------------------------------------------------
    return_type            Optional :class:`str`.
                           The return type of the expression in expressionInfo
    ==================     ====================================================================
    """

    def __init__(
        self, expression=None, name=None, title=None, return_type="boolean", **kwargs
    ):
        self._expression = expression
        self._name = name
        self._return_type = return_type
        self._title = title
        self._kwargs = kwargs

    def __repr__(self):
        if self._title:
            return "Expression: " + self._title
        else:
            return "Expression"

    @property
    def expression(self):
        """Gets/sets the expression for the expression info."""
        return self._expression

    @expression.setter
    def expression(self, value):
        if value:
            self._expression = value.replace('"', '"')
        else:
            raise ValueError("Expression must be set")

    @property
    def name(self):
        """Gets/sets the name for the expression info."""
        return self._name

    @name.setter
    def name(self, value):
        if value:
            self._name = value
        else:
            raise ValueError("Name must be set")

    @property
    def return_type(self):
        """Gets/sets the return type for the expression info."""
        return self._return_type

    @return_type.setter
    def return_type(self, value):
        self._return_type = value

    @property
    def title(self):
        """Gets/sets the title for the expression info."""
        return self._title

    @title.setter
    def title(self, value):
        self._title = value

    def to_dict(self):
        exp_dict = {}
        if self._name:
            exp_dict["name"] = self._name
        if self._title:
            exp_dict["title"] = self._title
        if self._expression:
            exp_dict["expression"] = self._expression
        if self._return_type:
            exp_dict["returnType"] = self._return_type
        for key, value in self._kwargs.items():
            exp_dict[key] = value
        return exp_dict



# Copyright 2018-2019 Open-MMLab.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# model settings
input_size = 300
model = dict(
    type="SingleStageDetector",
    pretrained="open-mmlab://vgg16_caffe",
    backbone=dict(
        type="SSDVGG",
        input_size=input_size,
        depth=16,
        with_last_pool=False,
        ceil_mode=True,
        out_indices=(3, 4),
        out_feature_indices=(22, 34),
        l2_norm_scale=20,
    ),
    neck=None,
    bbox_head=dict(
        type="SSDHead",
        in_channels=(512, 1024, 512, 256, 256, 256),
        num_classes=80,
        anchor_generator=dict(
            type="SSDAnchorGenerator",
            scale_major=False,
            input_size=input_size,
            basesize_ratio_range=(0.15, 0.9),
            strides=[8, 16, 32, 64, 100, 300],
            ratios=[[2], [2, 3], [2, 3], [2, 3], [2], [2]],
        ),
        bbox_coder=dict(
            type="DeltaXYWHBBoxCoder",
            target_means=[0.0, 0.0, 0.0, 0.0],
            target_stds=[0.1, 0.1, 0.2, 0.2],
        ),
    ),
    train_cfg=dict(
        assigner=dict(
            type="MaxIoUAssigner",
            pos_iou_thr=0.5,
            neg_iou_thr=0.5,
            min_pos_iou=0.0,
            ignore_iof_thr=-1,
            gt_max_assign_all=False,
        ),
        smoothl1_beta=1.0,
        allowed_border=-1,
        pos_weight=-1,
        neg_pos_ratio=3,
        debug=False,
    ),
    test_cfg=dict(
        nms=dict(type="nms", iou_threshold=0.45),
        min_bbox_size=0,
        score_thr=0.02,
        max_per_img=200,
    ),
)
cudnn_benchmark = True



from __future__ import annotations
from typing import Any, Optional, Union
from arcgis import env
from arcgis._impl.common._mixins import PropertyMap


########################################################################
class NetworkDiagramManager(object):
    """
    The Network Diagram service resource represents a network diagram
    service published with ArcGIS Server. The resource provides information
    about the service itself (name, type, default diagram template) and
    exposes various functions to access published network diagrams, create
    new network diagrams and store them, edit and maintain network diagrams, and so on.

    The Network Diagram service supports some operations which allow
    retrieving network diagrams, getting the characteristics of
    the diagrams you want (diagram info, consistency state), creating new network
    diagrams and deleting network diagrams.

    .. note::
        The active portal account must be licensed with the ArcGIS Utility Network
        user type extension or the ArcGIS Trace Network user type extension to
        use the utility network and network diagram services.

    =====================   ===========================================
    **Inputs**              **Description**
    ---------------------   -------------------------------------------
    url                     Required String. The web endpoint to the utility service.
    ---------------------   -------------------------------------------
    version                 Required Version. The `Version` class where the branch version will take place.
    ---------------------   -------------------------------------------
    gis                     Optional GIS. The `GIS` connection object.
    =====================   ===========================================


    """

    _con = None
    _gis = None
    _url = None
    _version = None
    _property = None
    _version_guid = None
    _version_name = None

    # ----------------------------------------------------------------------
    def __init__(self, url, version=None, gis=None):
        """Constructor"""
        if gis is None:
            gis = env.active_gis
        self._gis = gis
        self._con = gis._portal.con
        self._url = url
        self._version = version
        if version:
            self._version_guid = version._guid
            self._version_name = version.properties.versionName
        else:
            self._version_guid = None
            self._version_name = None

    # ----------------------------------------------------------------------
    def _init(self):
        """initializer"""
        try:
            res = self._con.get(self._url, {"f": "json"})
            self._property = PropertyMap(res)
        except Exception as e:
            self._property = PropertyMap({})

    # ----------------------------------------------------------------------
    @property
    def properties(self):
        """returns the properties for the service"""
        if self._property is None:
            self._init()
        return self._property

    # ----------------------------------------------------------------------
    def find_diagram_names(
        self,
        moment: Optional[int] = None,
        extent: Optional[dict] = None,
        where: Optional[str] = None,
        features: Optional[dict[str, Any]] = None,
        exclude_system_diagrams: bool = False,
    ):
        """
        The find_diagram_names operation is performed on a Network Diagram Service resource.
        The result of this operation is an array of strings, each one corresponding to
        a network diagram's name.

        This operation is used to retrieve the set of diagrams that cover
        a given extent, verify a particular WHERE clause, or contain specific
        utility network features or diagram features.

        ==============================      =====================================================
        **Parameter**                       **Description**
        ------------------------------      -----------------------------------------------------
        moment                              Optional integer. The session moment.
        ------------------------------      -----------------------------------------------------
        extent                              Optional dictionary representing the extent that you
                                            want the network extent of the resulting diagrams to
                                            intersect.

                                            Syntax:
                                            ```
                                            {
                                                "xmin": <xmin>,
                                                "ymin": <ymin>,
                                                "xmax": <xmax>,
                                                "ymax": <ymax>,
                                                "zmin": <zmin>,
                                                "zmax": <zmax>,
                                                "mmin": <mmin>,
                                                "mmax": <mmax>,
                                                "spatialReference": {<spatialReference>}
                                            }
                                            ```
        ------------------------------      -----------------------------------------------------
        where                               Optional string. Any legal SQL WHERE clause operating
                                            on some fields in the diagrams table is allowed. See
                                            table below for the exact list of field names that can
                                            be used.
        ------------------------------      -----------------------------------------------------
        features                            Optional dictionary. A set of utility network feature
                                            Global IDs, or network diagram feature Global IDs
                                            represented in a given diagram that are included in the
                                            resulting queried diagrams.

                                            The dictionary is composed of two keys:
                                            * `globalIDs` - An array of utility network feature
                                            Global IDs (case 1), or an array of network diagram
                                            feature Global IDs (case 2).
                                            * `diagram` - For case 1, NULL, or for case 2, the
                                            diagram name referencing the specified network diagram
                                            feature Global IDs.
        ------------------------------      -----------------------------------------------------
        exclude_system_diagrams             Optional bool. If True, the operation returns any
                                            diagrams except for the subnetwork system diagrams. If
                                            False (default), the operation returns any diagram.
        ==============================      =====================================================

        WHERE CLAUSE FIELDS:

        ================      ===================
        **Field**             **Type**
        ----------------      -------------------
        name                  String
        ----------------      -------------------
        tag                   String
        ----------------      -------------------
        creationDate          Date
        ----------------      -------------------
        creator               String
        ----------------      -------------------
        lastUpdateDate        Date
        ----------------      -------------------
        lastUpdateBy          String
        ================      ===================

        :return: An array of strings, each one corresponding to a diagram name.
        """
        url = "%s/findDiagramNames" % self._url
        params = {
            "f": "json",
            "gdbVersion": self._version_name,
            "sessionId": self._version_guid,
            "moment": moment,
            "extent": extent,
            "where": where,
            "features": features,
            "excludeSystemDiagrams": exclude_system_diagrams,
        }
        return self._con.post(url, params)

    # ----------------------------------------------------------------------
    def create_diagram_from_features(self, template: str, initial_features: list[str]):
        """
        The create_diagram_from_features operation is performed on a Network Diagram
        Service resource. The result of this operation is a Diagram Information JSON object.

        It is used to create a new temporary network diagram.

        ==============================      =====================================================
        **Parameter**                       **Description**
        ------------------------------      -----------------------------------------------------
        template                            Required string. The name of the diagram template the
                                            new network diagram will be based on.
        ------------------------------      -----------------------------------------------------
        initial_features                    Required list of strings. A list of utility network
                                            features Global IDs from which the new diagram is going
                                            to be built.
        ==============================      =====================================================

        :return: A dictionary of the diagram information.
        """
        url = "%s/createDiagramFromFeatures" % self._url
        params = {
            "f": "json",
            "gdbVersion": self._version_name,
            "sessionId": self._version_guid,
            "template": template,
            "initialFeatures": initial_features,
        }
        return self._con.post(url, params)

    # ----------------------------------------------------------------------
    def find_diagram_infos(
        self, diagrams_names: list[str], moment: Optional[int] = None
    ):
        """
        The find_diagram_infos operation is performed on a Network Diagram Service
        resource. The result of this operation is an array of Diagram Information JSON objects.

        ==============================      =====================================================
        **Parameter**                       **Description**
        ------------------------------      -----------------------------------------------------
        diagram_names                       Required list of strings. Each string corresponds to a
                                            diagram name for which you want to get diagram information.
        ------------------------------      -----------------------------------------------------
        moment                              Optional Integer. A session moment.
        ==============================      =====================================================


        :return: Diagram info object for each of the diagram names specified in input.
        """
        url = "%s/findDiagramInfos" % self._url
        params = {
            "f": "json",
            "gdbVersion": self._version_name,
            "sessionId": self._version_guid,
            "diagramNames": diagrams_names,
            "moment": moment,
        }
        return self._con.post(url, params)

    # ----------------------------------------------------------------------
    def query_consistency_state(
        self, diagrams_names: list[str], moment: Optional[int] = None
    ):
        """
        The query_consistency_state operation is performed on a Network Diagram Service resource.

        ==============================      =====================================================
        **Parameter**                       **Description**
        ------------------------------      -----------------------------------------------------
        diagram_names                       Required list of strings. Each string corresponds to a
                                            diagram name for which you want to get diagram information.
        ------------------------------      -----------------------------------------------------
        moment                              Optional Integer. A session moment.
        ==============================      =====================================================

        :return: It returns the consistency state for each of the diagram names specified in input.
        """
        url = "%s/queryConsistencyState" % self._url
        params = {
            "f": "json",
            "gdbVersion": self._version_name,
            "sessionId": self._version_guid,
            "diagramNames": diagrams_names,
            "moment": moment,
        }
        return self._con.post(url, params)

    # ----------------------------------------------------------------------
    def delete_diagram(self, name: str):
        """
        The delete_diagram operation is performed on a Network Diagram Service
        resource.

        ==============================      =====================================================
        **Parameter**                       **Description**
        ------------------------------      -----------------------------------------------------
        name                                Required string. The name of the network diagram to delete.
        ==============================      =====================================================

        :return: The moment (date) the delete_diagram operation happens.

        """

        url = "%s/deleteDiagram" % self._url
        params = {
            "f": "json",
            "gdbVersion": self._version_name,
            "sessionId": self._version_guid,
            "name": name,
        }
        return self._con.post(url, params)

    # ----------------------------------------------------------------------
    @property
    def templates(self):
        """
        The Diagram Templates resource represents all the diagram templates
        under a Network Diagram service. It returns an array of template names.
        """
        url = "%s/templates" % self._url
        params = {
            "f": "json",
            "gdbVersion": self._version_name,
            "sessionId": self._version_guid,
        }
        return self._con.post(url, params)

    # ----------------------------------------------------------------------
    def template(self, name):
        """
        The Template resource represents a single diagram template
        under a Network Diagram service. It is returned as a JSON
        object that provides the list of the diagram layouts and their
        property sets, which are preset for the template.

        The Template resource doesn't support any operation nor any child resource.

        ==============================      =====================================================
        **Parameter**                       **Description**
        ------------------------------      -----------------------------------------------------
        name                                Required string. The name of the template.
        ==============================      =====================================================

        """
        url = "%s/templates/%s" % (self._url, name)
        params = {
            "f": "json",
            "gdbVersion": self._version_name,
            "sessionId": self._version_guid,
        }
        return self._con.post(url, params)

    # ----------------------------------------------------------------------
    def diagrams(self, moment: Optional[int] = None):
        """
        The Network Diagrams resource represents all the network diagrams
        under a Network Diagram service. It is returned as an array of network
        diagram names.
        By default, only the diagrams in Default are returned.

        ==============================      =====================================================
        **Parameter**                       **Description**
        ------------------------------      -----------------------------------------------------
        moment                              Optional int. The session moment.
        ==============================      =====================================================

        :return: A list of diagram names.
        """

        url = "%s/diagrams" % self._url
        params = {
            "f": "json",
            "gdbVersion": self._version_name,
            "sessionId": self._version_guid,
            "moment": moment,
        }
        return self._con.post(url, params)

    # ----------------------------------------------------------------------
    def diagram(self, name: str):
        """
        The Diagram resource represents a single network diagram under
        a Network Diagram service.

        ==============================      =====================================================
        **Parameter**                       **Description**
        ------------------------------      -----------------------------------------------------
        name                                Required string. The name of the network diagram.
        ==============================      =====================================================

        """
        url = "%s/diagrams/%s" % (self._url, name)
        return Diagram(url, self._version, self._gis)

    # ----------------------------------------------------------------------
    @property
    def diagram_dataset(self):
        """
        The Diagram Dataset resource regroups the info related to each diagram
        template under a Network Diagram service. It returns an array of diagram template info.
        """
        url = "%s/diagramDataset" % self._url
        params = {
            "f": "json",
            "gdbVersion": self._version_name,
            "sessionId": self._version_guid,
        }
        return self._con.post(url, params)


# --------------------------------------------------------------------------
class Diagram(object):
    """
    The Diagram resource represents a single network diagram under a Network
    Diagram service. It is returned as a JSON Diagram Information object.

    It supports three child resources:
    * Diagram Map—Mimics a map service resource for the network diagram.
    * Dynamic Layers—Describes the sublayers under the diagram layer.
    * Layer Definitions—Details the layer and labeling properties that define each sublayer under the diagram layer.
    """

    _con = None
    _gis = None
    _url = None
    _version = None
    _property = None
    _version_guid = None
    _version_name = None

    # ----------------------------------------------------------------------
    def __init__(self, url, version=None, gis=None):
        """Constructor"""
        if gis is None:
            gis = env.active_gis
        self._gis = gis
        self._con = gis._portal.con
        self._url = url
        self._version = version
        if version:
            self._version_guid = version._guid
            self._version_name = version.properties.versionName
        else:
            self._version_guid = None
            self._version_name = None

    # ----------------------------------------------------------------------
    def _init(self):
        """initializer"""
        try:
            res = self._con.get(
                self._url,
                {
                    "f": "json",
                    "gdbVersion": self._version_name,
                    "sessionId": self._version_guid,
                },
            )
            self._property = PropertyMap(res)
        except Exception as e:
            self._property = PropertyMap({})

    # ----------------------------------------------------------------------
    @property
    def properties(self):
        """returns the properties for the service"""
        if self._property is None:
            self._init()
        return self._property

    # ----------------------------------------------------------------------
    def append_features(self, added_features: list[str]):
        """
        The append_features operation is performed on a Diagram resource.
        The result of this operation is a Diagram JSON Information object,
        and the moment (date) the appendFeatures operation
        happens for a stored diagram.

        It is used to append a set of utility network feature to the diagram resource.

        ==============================      =====================================================
        **Parameter**                       **Description**
        ------------------------------      -----------------------------------------------------
        added_features                      Required list of strings. The strings are utility network
                                            feature Global IDs, the features being appended to the
                                            diagram resource.
        ==============================      =====================================================
        """

        url = "%s/appendFeatures" % self._url
        params = {
            "f": "json",
            "gdbVersion": self._version_name,
            "sessionId": self._version_guid,
            "addedFeatures": added_features,
        }
        return self._con.post(url, params)

    # ----------------------------------------------------------------------
    def apply_layout(
        self,
        layout_name: str,
        layer_params: Optional[dict[str, Any]] = None,
        junction_ids: list = None,
        container_ids: list = None,
        edge_ids: list = None,
        run_async: bool = False,
    ):
        """
        The apply_layout operation is performed on a Diagram resource.
        The result of this operation is a Diagram JSON Information object,
        and the moment (date) the applyLayout operation happens for a stored diagram.

        It is used to apply a specific diagram algorithm on all or parts of
        the resource diagram content.

        ==============================      =====================================================
        **Parameter**                       **Description**
        ------------------------------      -----------------------------------------------------
        layout_name                         Required string. The name of the algorithm layout to
                                            execute.

                                            `Values: "AngleDirectedDiagramLayout" | "CompressionDiagramLayout" |
                                            "ForceDirectedDiagramLayout" | "GeoPositionsDiagramLayout" |
                                            "GridDiagramLayout" | "LinearDispatchDiagramLayout" |
                                            "MainLineTreeDiagramLayout" | "MainRingDiagramLayout" |
                                            "PartialOverlappingEdgesDiagramLayout" | "RadialTreeDiagramLayout" |
                                            "RelativeMainlineDiagramLayout" | "ReshapeEdgesDiagramLayout" |
                                            "RotateTreeDiagramLayout" | "SeparateOverlappingEdgesDiagramLayout" |
                                            "SmartTreeDiagramLayout" | "SpatialDispatchDiagramLayout"`
        ------------------------------      -----------------------------------------------------
        layout_params                       Optional dictionary. The algorithm layer parameters.

                                            `Specific Property Parameters <https://developers.arcgis.com/rest/services-reference/enterprise/appendix-diagram-layout-property-set-objects.htm\>`_

                                            Example:
                                               layoutParams = {
                                                   "type": "PropertySet",
                                                    "propertySetItems": [
                                                        "are_containers_preserved",
                                                        false,
                                                        "is_active",
                                                        false,
                                                        "iterations_number",
                                                        20,
                                                        "repel_factor",
                                                        1,
                                                        "degree_freedom",
                                                        1
                                                    ]
                                               }
        ------------------------------      -----------------------------------------------------
        junction_ids                        Optional list. For the case you want the layout algorithm
                                            to execute on a diagram part. A list of junction object Ids
                                            that will be processed.
        ------------------------------      -----------------------------------------------------
        container_ids                       Optional list. For the case you want the layout algorithm
                                            to execute on a diagram part. A list of container object Ids
                                            that will be processed.
        ------------------------------      -----------------------------------------------------
        edge_ids                            Optional list. For the case you want the layout algorithm
                                            to execute on a diagram part. A list of edge object Ids
                                            that will be processed.
        ------------------------------      -----------------------------------------------------
        run_async                           Optional bool. Specify whether the layout algorithm will
                                            run synchronously or asynchronously.

                                            If False, the layout algorithm will run synchronously
                                            and can fail if its execution exceeds the service timeout—600
                                            seconds by default. This is the default.

                                            If True, the layout algorithm will run asynchronously.
                                            This option dedicates server resources to run the layout
                                            algorithm with a longer time-out. Running asynchronously can
                                            be interesting when executing layout that are time
                                            consuming—for example, Partial Overlapping Edges—and
                                            applying to large diagrams—more than 25,000 features.
        ==============================      =====================================================

        """
        url = "%s/applyLayout" % self._url
        params = {
            "f": "json",
            "gdbVersion": self._version_name,
            "sessionId": self._version_guid,
            "layoutName": layout_name,
            "layoutParams": layer_params,
            "junctionObjectIDs": junction_ids,
            "containerObjectIDs": container_ids,
            "edgeObjectIDs": edge_ids,
        }
        if self._gis.version >= [9, 2]:
            params["async"] = run_async
        return self._con.post(url, params)

    # ----------------------------------------------------------------------
    def apply_template_layouts(
        self,
        junction_ids: Optional[list] = None,
        container_ids: Optional[list] = None,
        edge_ids: Optional[list] = None,
    ):
        """
        The apply_template_layouts operation is performed on a Diagram resource.
        The result of this operation is a Diagram JSON Information object, and
        the moment the applyTemplateLayouts operation happens for a stored diagram.

        It is used to re-execute the list of the layout algorithms currently configured
        on the template the resource diagram is based on.

        ==============================      =====================================================
        **Parameter**                       **Description**
        ------------------------------      -----------------------------------------------------
        junction_ids                        Optional list. For the case you want the layout algorithm
                                            to execute on a diagram part. A list of junction object Ids
                                            that will be processed.
        ------------------------------      -----------------------------------------------------
        container_ids                       Optional list. For the case you want the layout algorithm
                                            to execute on a diagram part. A list of container object Ids
                                            that will be processed.
        ------------------------------      -----------------------------------------------------
        edge_ids                            Optional list. For the case you want the layout algorithm
                                            to execute on a diagram part. A list of edge object Ids
                                            that will be processed.
        ==============================      =====================================================
        """
        url = "%s/applyTemplateLayouts" % self._url
        params = {
            "f": "json",
            "gdbVersion": self._version_name,
            "sessionId": self._version_guid,
            "junctionObjectIDs": junction_ids,
            "containerObjectIDs": container_ids,
            "edgeObjectIDs": edge_ids,
        }
        return self._con.post(url, params)

    # ----------------------------------------------------------------------
    def clear_flags(self, flag_type: str):
        """
        The clear_flags operation is performed on a Diagram resource.
        This operation returns the moment (date) the clear_flags operation happens
        when it applies on a stored diagram.

        It is used to clear all root junction, end junction, pivot
        junction and barrier flags on the resource diagram.

        ==============================      =====================================================
        **Parameter**                       **Description**
        ------------------------------      -----------------------------------------------------
        flag_type                           Required string. The type of flag you want to clear in
                                            the diagram.

                                            `Values: "esriDiagramRootJunction" | "esriDiagramEndJunction" |
                                            "esriDiagramPivotJunction" | "esriDiagramBarrierEdge" |
                                            "esriDiagramBarrierJunction"`
        ==============================      =====================================================
        """
        url = "%s/clearFlags" % self._url
        params = {
            "f": "json",
            "gdbVersion": self._version_name,
            "sessionId": self._version_guid,
            "flagType": flag_type,
        }
        return self._con.post(url, params)

    # ----------------------------------------------------------------------
    def extend(self, extend_type: str = None, from_features: list[str] = None):
        """
        The extend operation is performed on a Diagram resource.
        The result of this operation is a Diagram JSON Information object,
        and the moment the edit operation happens for a stored diagram.

        It is used to extend the diagram resource content one connectivity level,
        optionally regarding to the traversability over the network.

        ==============================      =====================================================
        **Parameter**                       **Description**
        ------------------------------      -----------------------------------------------------
        extend_type                         Optional string. The type of extend you want to process.

                                            `Values: "esriDiagramExtendByAttachment" |
                                            "esriDiagramExtendByConnectivity" |
                                            "esriDiagramExtendByTraversability" |
                                            "esriDiagramExtendByContainment"`
        ------------------------------      -----------------------------------------------------
        from_features                       Optional list of strings. Diagram feature Global IDs,
                                            those diagram features being those from which the
                                            extend process will execute.
        ==============================      =====================================================
        """
        url = "%s/extend" % self._url
        params = {
            "f": "json",
            "gdbVersion": self._version_name,
            "sessionId": self._version_guid,
            "extendType": extend_type,
            "fromFeatures": from_features,
        }
        return self._con.post(url, params)

    # ----------------------------------------------------------------------
    def find_diagram_features(
        self,
        from_features: list[str],
        include_aggregations: bool,
        add_connectivity_associations: bool,
        add_structural_attachments: bool,
        from_diagram: Optional[str] = None,
        moment: Optional[int] = None,
    ):
        """
        The find_diagram_features operation is performed on a Diagram resource.
        It returns an array of diagram feature JSON objects for the input fromDiagram.

        This is the operation to use to search for the utility network features
        associated with a set of diagram features that are referenced in the diagram resource.

        ==============================      =====================================================
        **Parameter**                       **Description**
        ------------------------------      -----------------------------------------------------
        from_features                       Required list of strings. Depending on whether you
                                            want to retrieve diagram features associated with
                                            utility network features (case#1) or with diagram
                                            features that are represented in another diagram (case#2),
                                            an array of utility network feature Global IDs (case#1),
                                            or an array of network diagram feature Global IDs (case#2).
        ------------------------------      -----------------------------------------------------
        include_aggregations                Required boolean.
                                            Case#1—When the fromFeatures reference utility
                                            network feature Global IDs:
                                            * `True`—The operation returns the diagram features that
                                            strictly represent those utility network features in the
                                            diagram resource, and the diagram features that are associated
                                            with those utility network features but not represented in
                                            the diagram resource where they are reduced or collapsed.
                                            * `False`—The operation only returns the diagram features
                                            associated with those utility network features that are
                                            not reduced nor collapsed in the diagram resource; that is,
                                            it only returns the diagram features associated with those
                                            utility network features that are visibly represented
                                            in the diagram resource.


                                            Case#2—When the fromFeatures reference diagram feature
                                            Global IDs represented in another diagram:
                                            * `True`—The operation returns the diagram features
                                            associated with the same utility network features
                                            those diagram features are, whether those features
                                            are reduced or collapsed in this other diagram and/or
                                            in the resource diagram.
                                            * `False`—The operation only returns the diagram features
                                            associated with the same utility network features that
                                            are visibly represented in this other diagram
                                            and in the resource diagram.
        ------------------------------      -----------------------------------------------------
        add_connectivity_associations       Required boolean. When the from_features reference utility
                                            network feature Global IDs:

                                            * `True`—The operation also adds any connectivity association
                                            diagram edges for which it has just retrieved both the "from"
                                            and "to" diagram junctions.
                                            * `False`—The operation doesn't add any connectivity
                                            association diagram edges represented in the diagram resource.
        ------------------------------      -----------------------------------------------------
        add_structural_attachments          Required boolean. When the from_features reference
                                            utility network feature Global IDs:

                                            * `True`—The operation also adds any structural attachment
                                            diagram edges for which it has just retrieved both the
                                            "from" and "to" diagram junctions.
                                            * `False`—The operation doesn't add any structural
                                            attachment diagram edges represented in the
                                            diagram resource.
        ------------------------------      -----------------------------------------------------
        from_diagram                        Optional string. The name of the diagram the feature
                                            Global IDs specified for the from_features parameters
                                            belong to, or null when the from_features parameter
                                            references utility network features.
        ------------------------------      -----------------------------------------------------
        moment                              Optional integer. The session moment.
        ==============================      =====================================================
        """
        url = "%s/findDiagramFeatures" % self._url
        params = {
            "f": "json",
            "gdbVersion": self._version_name,
            "sessionId": self._version_guid,
            "moment": moment,
            "fromFeatures": from_features,
            "fromDiagram": from_diagram,
            "includeAggregations": include_aggregations,
            "addConnectivityAssociations": add_connectivity_associations,
            "addStructuralAttachments": add_structural_attachments,
        }
        return self._con.post(url, params)

    # ----------------------------------------------------------------------
    def find_initial_network_objects(self, moment: Optional[int] = None):
        """
        The find_initial_network_objects operation is performed on a Diagram resource.
        It returns an array of network feature globalIDs.

        This is the operation to use to search for the set of network
        features used as input for the diagram build.

        ==============================      =====================================================
        **Parameter**                       **Description**
        ------------------------------      -----------------------------------------------------
        moment                              Optional integer. A session moment.
        ==============================      =====================================================
        """
        url = "%s/findInitialNetworkObjects" % self._url
        params = {
            "f": "json",
            "gdbVersion": self._version_name,
            "sessionId": self._version_guid,
            "moment": moment,
        }
        return self._con.post(url, params)

    # ----------------------------------------------------------------------
    def find_network_features(
        self,
        from_features: list[str],
        include_aggregations: bool,
        moment: Optional[int] = None,
    ):
        """
        The find_network_features operation is performed on a Diagram resource.
        It returns an array of network feature global IDs.

        This is the operation to use to search for the diagram features referenced
        in the diagram resource that are associated with a set of utility
        network features or a set of diagram features
        represented in another network diagram.

        ==============================      =====================================================
        **Parameter**                       **Description**
        ------------------------------      -----------------------------------------------------
        from_features                       Required list of strings. A list of diagram network
                                            feature global IDs.
        ------------------------------      -----------------------------------------------------
        include_aggregations                Required boolean.
                                            * `True`—The operation returns all the network features
                                            associated with the diagram features specified in the
                                            from_features parameter, whether those features
                                            are reduced or collapsed in the diagram resource.
                                            * `False`—The operation only returns the network features
                                            associated with diagram features specified in the
                                            from_features parameter that are not reduced or
                                            collapsed in the diagram resource; that is, it only
                                            returns the network features associated with the specified
                                            diagram features that are visibly represented in the diagrams.
        ------------------------------      -----------------------------------------------------
        moment                              Optional integer. A session moment.
        ==============================      =====================================================

        """
        url = "%s/findNetworkFeatures" % self._url
        params = {
            "f": "json",
            "gdbVersion": self._version_name,
            "sessionId": self._version_guid,
            "moment": moment,
            "fromFeatures": from_features,
            "includeAggregations": include_aggregations,
        }
        return self._con.post(url, params)

    # ----------------------------------------------------------------------
    def get_aggregations(self, moment: Optional[int] = None):
        """
        The get_aggregations operation is performed on a Diagram resource.
        The result of this operation is an array of Diagram Aggregation JSON objects.

        It returns all the diagram aggregations in the diagram resource.

        ==============================      =====================================================
        **Parameter**                       **Description**
        ------------------------------      -----------------------------------------------------
        moment                              Optional integer. A session moment.
        ==============================      =====================================================
        """
        url = "%s/getAggregations" % self._url
        params = {
            "f": "json",
            "gdbVersion": self._version_name,
            "sessionId": self._version_guid,
            "moment": moment,
        }
        return self._con.post(url, params)

    # ----------------------------------------------------------------------
    def get_flags(
        self,
        flag_type: str,
        moment: Optional[int] = None,
        out_sr: Optional[Union[str, dict]] = None,
    ):
        """
        The get_flags operation is performed on a Diagram resource.
        The result of this operation is a JSON Information object that
        returns the list of diagram element IDs bringing a flag, with its
        flag type and its location.

        It is used to get the root junction, end junction, pivot junction,
        and barrier flag on a particular diagram feature.

        ==============================      =====================================================
        **Parameter**                       **Description**
        ------------------------------      -----------------------------------------------------
        flag_type                           Required string. The type of flag you want to search
                                            for on the diagram resource.

                                            `Values: "esriDiagramRootJunction" | "esriDiagramEndJunction" |
                                            "esriDiagramPivotJunction" | "esriDiagramBarrierEdge" |
                                            "esriDiagramBarrierJunction"`
        ------------------------------      -----------------------------------------------------
        moment                              Optional integer. A session moment.
        ------------------------------      -----------------------------------------------------
        out_sr                              Optional dictionary or string to specify the spatial
                                            reference of the returned geometry.
        ==============================      =====================================================
        """
        url = "%s/getFlags" % self._url
        params = {
            "f": "json",
            "gdbVersion": self._version_name,
            "sessionId": self._version_guid,
            "moment": moment,
            "flagType": flag_type,
            "outSR": out_sr,
        }
        return self._con.post(url, params)

    # ----------------------------------------------------------------------
    def manage_flag(self, flag_type: str, flag_id: str, action: str):
        """
        The manage_flag operation is performed on a Diagram resource.
        This operation returns the moment the manageFlag operation happens
        when it applies on a stored diagram.

        It is used to add or remove root junction, end junction, pivot
        junction and barrier flags on a particular diagram feature.


        ==============================      =====================================================
        **Parameter**                       **Description**
        ------------------------------      -----------------------------------------------------
        flag_type                           Required string. The type of flag you want to search
                                            for on the diagram resource.

                                            `Values: "esriDiagramRootJunction" | "esriDiagramEndJunction" |
                                            "esriDiagramPivotJunction" | "esriDiagramBarrierEdge" |
                                            "esriDiagramBarrierJunction"`
        ------------------------------      -----------------------------------------------------
        flag_id                             Required string. The diagram element id of the diagram
                                            feature to which you want the flag to be added or from
                                            which you want the flag to be removed.
        ------------------------------      -----------------------------------------------------
        action                              Required string. The flag operation to execute:
                                            * "add" - to add a new flag on a particular diagram feature.
                                            * "remove" - to remove a flag from a particular
                                            diagram feature.
        ==============================      =====================================================
        """
        url = "%s/manageFlag" % self._url
        params = {
            "f": "json",
            "gdbVersion": self._version_name,
            "sessionId": self._version_guid,
            "flagID": flag_id,
            "flagType": flag_type,
            "action": action,
        }
        return self._con.post(url, params)

    # ----------------------------------------------------------------------
    def overwrite_from_features(self, initial_features: list[str]):
        """
        The overwrite_from_features operation is performed on a Diagram resource.
        The result of this operation is a Diagram JSON Information object,
        and the moment (date) the overwriteFromFeatures operation happens for a stored diagram.

        It is used to overwrite the diagram resource content from a set of
        utility network feature Global IDs.

        ==============================      =====================================================
        **Parameter**                       **Description**
        ------------------------------      -----------------------------------------------------
        initial_features                    Required list of strings. A list of utility network
                                            feature Global IDs, those features being the ones used
                                            to overwrite the diagram resource.
        ==============================      =====================================================
        """
        url = "%s/overwriteFromFeatures" % self._url
        params = {
            "f": "json",
            "gdbVersion": self._version_name,
            "sessionId": self._version_guid,
            "initialFeatures": initial_features,
        }
        return self._con.post(url, params)

    # ----------------------------------------------------------------------
    def query_diagram_content(
        self,
        moment: Optional[int] = None,
        add_diagram_info: bool = False,
        add_geometries: bool = False,
        add_attributes: bool = False,
        add_aggregations: bool = False,
        use_value_names: bool = False,
        out_sr: Optional[Union[str, dict]] = None,
    ):
        """
        The query_diagram_content operation is performed on a diagram resource.
        It returns the diagram content in a simple format that reflects basic
        connectivity. Other optional information can also be returned,
        such as diagram feature geometry, network element attributes with their
        string descriptions rather than raw values for coded domain values, aggregated
        elements, and diagram properties.

        ==============================      =====================================================
        **Parameter**                       **Description**
        ------------------------------      -----------------------------------------------------
        moment                              Optional int. The session moment.
        ------------------------------      -----------------------------------------------------
        add_diagram_info                    Optional bool.
                                            * `True`—The operation also returns some extra
                                            information on the diagram, such as its template,
                                            the diagram statistics, its creation and last update dates,
                                            and the diagram extent.
                                            * `False`—The operation doesn't return extra
                                            information on the diagram (default).
        ------------------------------      -----------------------------------------------------
        add_geometries                      Optional bool.
                                            * `True`—The operation returns each diagram feature with its geometry.
                                            * `False`—The operation doesn't return the diagram feature geometries (default).
        ------------------------------      -----------------------------------------------------
        add_attributes                      Optional bool.
                                            * `True`—The operation also returns the attributes
                                            of the utility network feature associated with
                                            each diagram feature.
                                            * `False`—The operation doesn't
                                            return any attributes (default).
        ------------------------------      -----------------------------------------------------
        add_aggregations                    Optional bool.
                                            * `True`—The operation returns each diagram feature
                                            with the list of the utility network elements it
                                            aggregates with their asset group and asset type values.
                                            * `False`—The operation doesn't return any aggregations (default)
        ------------------------------      -----------------------------------------------------
        use_value_names                     Optional bool.
                                            For the cases in which the associated network feature
                                            attributes or aggregations are exported, that is,
                                            `add_attributes = True` or `add_aggregations = True`,
                                            the method to use to export coded domain and subtype values:

                                            * `True`—Coded domain and subtype values will be
                                            exported using their string descriptions rather than raw values.
                                            * `False`—Coded domain and subtype values will be
                                            exported as raw values. This is the default.
        ------------------------------      -----------------------------------------------------
        out_sr                              Optional dictionary or string. The spatial reference of the
                                            returned geometry.
        ==============================      =====================================================

        """
        url = "%s/queryDiagramContent" % self._url
        params = {
            "f": "json",
            "gdbVersion": self._version_name,
            "sessionId": self._version_guid,
            "moment": moment,
            "addDiagramInfo": add_diagram_info,
            "addGeometries": add_geometries,
            "addAttributes": add_attributes,
            "addAggregations": add_aggregations,
            "useValueNames": use_value_names,
            "outSR": out_sr,
        }
        return self._con.post(url, params)

    # ----------------------------------------------------------------------
    def query_elements_by_extent(
        self,
        extent: Optional[dict] = None,
        moment: Optional[int] = None,
        add_contents: bool = False,
        return_junctions: bool = True,
        return_edges: bool = True,
        return_containers: bool = True,
        return_geometry: bool = True,
        out_sr: Optional[Union[str, dict]] = None,
    ):
        """
        The query_elements_by_extent operation is performed on a diagram resource.
        The result of this operation is a JSON object composed of three arrays—one
        for the resulting diagram junctions, another for the resulting diagram edges,
        and the third one for the diagram containers.

        It returns the set of diagram features (diagram junctions, diagram edges,
        or diagram containers) represented in the diagram resource that intersects
        a specified envelope—this one being optionally enlarged to include any
        containers that may be partially within it.

        ==============================      =====================================================
        **Parameter**                       **Description**
        ------------------------------      -----------------------------------------------------
        extent                              Optional dictionary. Represent the extent you want the
                                            resulting diagram features to intersect.

                                            Without specifying this parameter, the operation returns
                                            all the diagram features in the resource diagram.

                                            .. code-block:: python

                                                {
                                                    "xmin": <xmin>,
                                                    "ymin": <ymin>,
                                                    "xmax": <xmax>,
                                                    "ymax": <ymax>,
                                                    "zmin": <zmin>,
                                                    "zmax": <zmax>,
                                                    "mmin": <mmin>,
                                                    "mmax": <mmax>,
                                                    "spatialReference": <spatialReference>
                                                }
        ------------------------------      -----------------------------------------------------
        moment                              Optional int. The session moment.
        ------------------------------      -----------------------------------------------------
        add_contents                        Optional bool.
                                            * `False` —To return the diagram features which strictly
                                            intersect the specified envelope (default)
                                            * `True`—To enlarge the searching envelope so it
                                            includes the extent of any containers that are partially
                                            within the specified envelope
        ------------------------------      -----------------------------------------------------
        return_junctions                    Optional bool.
                                            * `True`—To return the diagram junctions (default)
                                            * `False`—To not return the diagram junctions
        ------------------------------      -----------------------------------------------------
        return_edges                        Optional bool.
                                            * `True`—To return the diagram edges (default)
                                            * `False`—To not return the edges junctions
        ------------------------------      -----------------------------------------------------
        return_containers                   Optional bool.
                                            * `True`—To return the diagram containers (default)
                                            * `False`—To not return the diagram containers
        ------------------------------      -----------------------------------------------------
        return_geometry                     Optional bool.
                                            * `True`—To return each queried diagram feature with its
                                            geometry (default)
                                            * `False`—To not return the geometry.
        ------------------------------      -----------------------------------------------------
        out_sr                              Optional dictionary or string representing the spatial
                                            reference of the returned geometry.
        ==============================      =====================================================
        """
        url = "%s/queryDiagramElementsByExtent" % self._url
        params = {
            "f": "json",
            "gdbVersion": self._version_name,
            "sessionId": self._version_guid,
            "extent": extent,
            "moment": moment,
            "addContents": add_contents,
            "returnJunctions": return_junctions,
            "returnEdges": return_edges,
            "returnContainers": return_containers,
            "returnGeometry": return_geometry,
            "outSR": out_sr,
        }
        return self._con.post(url, params)

    # ----------------------------------------------------------------------
    def query_elements_by_ids(
        self,
        junction_ids: Optional[list] = None,
        container_ids: Optional[list] = None,
        edge_ids: Optional[list] = None,
        moment: Optional[int] = None,
        add_connected: bool = False,
        return_geometry: bool = True,
        out_sr: Optional[Union[dict, str]] = None,
    ):
        """
        The query_elements_by_ids operation is performed on a diagram resource.
        The result of this operation is a JSON object composed of three
        arrays—one for the resulting diagram junctions, another for the
        resulting diagram edges, and the third one for the diagram containers.

        It returns the set of diagram features—that is, diagram junctions,
        diagram edges, or diagram containers—represented in the diagram
        resource with the specified ObjectIDs. Then when specifying diagram edge
        ObjectIDs, optionally, extend the resulting set of diagram elements to the
        junctions they are connected to or when specifying diagram container ObjectIDs,
        optionally, extend the resulting set of diagram elements to the edge, junction,
        or container diagram elements they contain.

        ==============================      =====================================================
        **Parameter**                       **Description**
        ------------------------------      -----------------------------------------------------
        junction_ids                        Optional list. For the case you want the layout algorithm
                                            to execute on a diagram part. A list of junction object Ids
                                            that will be processed.
        ------------------------------      -----------------------------------------------------
        container_ids                       Optional list. For the case you want the layout algorithm
                                            to execute on a diagram part. A list of container object Ids
                                            that will be processed.
        ------------------------------      -----------------------------------------------------
        edge_ids                            Optional list. For the case you want the layout algorithm
                                            to execute on a diagram part. A list of edge object Ids
                                            that will be processed.
        ------------------------------      -----------------------------------------------------
        moment                              Optional integer. The session moment.
        ------------------------------      -----------------------------------------------------
        add_connected                       Optional bool.
                                            * `False`-Doesn't enlarge the returned diagram junction to
                                            those that are from or to junctions of the specified
                                            edge_ids. (default)
                                            * `True`-Enlarges the returned diagram junctions to those
                                            that are from or to junctions of the specifies edge_ids.
        ------------------------------      -----------------------------------------------------
        return_geometry                     Optional bool.
                                            * `False`-Return each queried feature without its geometry.
                                            * `True`-Return each queried feature with its geometry. (default)
        ------------------------------      -----------------------------------------------------
        out_sr                              Optional dictionary or string representing the spatial
                                            reference for the returned geometry.
        ==============================      =====================================================
        """
        url = "%s/queryDiagramElementsByObjectIDs" % self._url
        params = {
            "f": "json",
            "gdbVersion": self._version_name,
            "sessionId": self._version_guid,
            "junctionObjectIDs": junction_ids,
            "containerObjectIDs": container_ids,
            "edgeObjectIDs": edge_ids,
            "moment": moment,
            "addConnected": add_connected,
            "returnGeometry": return_geometry,
            "outSR": out_sr,
        }
        return self._con.post(url, params)

    # ----------------------------------------------------------------------
    def save_layout(
        self,
        junctions: Optional[list[dict]] = None,
        containers: Optional[list[dict]] = None,
        edges: Optional[list[dict]] = None,
    ):
        """
        The save_layout operation is performed on a Diagram resource.
        The result of this operation is a Diagram JSON Information object,
        and the moment (date) the saveLayout operation happens for a stored diagram.

        It is used to apply and save new geometries that may have been computed for a
        set of diagram features represented in the diagram resource.

        ==============================      =====================================================
        **Parameter**                       **Description**
        ------------------------------      -----------------------------------------------------
        junctions                           Optional list of dictionaries. Provide the new geometry
                                            for each edited junction, those diagram junctions being
                                            identified thanks to their Diagram Element ID.

                                            Syntax:
                                            junctions=[{"ID": <jctDEID1>, "geometry": <geometry1>}, ...]
        ------------------------------      -----------------------------------------------------
        containers                          Optional list of dictionaries. Provide the new geometry
                                            for each edited container, those diagram containers being
                                            identified thanks to their Diagram Element ID.

                                            Syntax:
                                            containers=[{"ID": <containerDEID1>, "geometry": <geometry1>}, ...]
        ------------------------------      -----------------------------------------------------
        edges                               Optional list of dictionaries. Provide the new geometry
                                            for each edited edge, those diagram edges being
                                            identified thanks to their Diagram Element ID.

                                            Syntax:
                                            containers=[{"ID": <edgeDEID1>, "geometry": <geometry1>}, ...]
        ==============================      =====================================================
        """
        url = "%s/saveLayout" % self._url
        params = {
            "f": "json",
            "gdbVersion": self._version_name,
            "sessionId": self._version_guid,
            "junctions": junctions,
            "containers": containers,
            "edges": edges,
        }
        return self._con.post(url, params)

    # ----------------------------------------------------------------------
    def store(self, name: str, access: str):
        """
        The store operation is performed on a Diagram resource.
        The result of this operation is a Diagram JSON Information object,
        and the moment the store operation happens.

        It is used to store the temporary network diagram resource in the database.

        ==============================      =====================================================
        **Parameter**                       **Description**
        ------------------------------      -----------------------------------------------------
        name                                Required string. The name of the network diagram to
                                            be stored.
        ------------------------------      -----------------------------------------------------
        access                              Required string. The access right level you want to
                                            set for the stored diagram.

                                            Values:
                                            * "esriDiagramPublicAccess" - Anyone will have full
                                            access rights on the diagram; that is,
                                            view/edit/update/overwrite/delete permissions on it.
                                            * "esriDiagramProtectedAccess" - Anyone will have view
                                            and read access rights on the diagram, but they will have
                                            no permission for editing/update/overwriting/nor
                                            deleting the diagram.
                                            * "esriDiagramPrivateAccess" - No access to the
                                            diagram—nor to view nor to edit it—for anyone except its owner
        ==============================      =====================================================
        """
        url = "%s/store" % self._url
        params = {
            "f": "json",
            "gdbVersion": self._version_name,
            "sessionId": self._version_guid,
            "name": name,
            "access": access,
        }
        return self._con.post(url, params)

    # ----------------------------------------------------------------------
    def update(self):
        """
        The update operation is performed on a Diagram resource.
        The result of this operation is a Diagram JSON Information object,
        and the moment the update operation happens for a stored diagram.

        It is used to update the diagram resource content; that is, synchronize
        its content from the network features used to initially generate it,
        and so reflect any changes that may have impacted those network
        features into the diagram.

        """
        url = "%s/update" % self._url
        params = {
            "f": "json",
            "gdbVersion": self._version_name,
            "sessionId": self._version_guid,
        }
        return self._con.post(url, params)

    # ----------------------------------------------------------------------
    def dynamic_layers(self, all_layers: bool = False):
        """
        The Dynamic Layers resource describes the sublayers under the diagram layer.

        ==============================      =====================================================
        **Parameter**                       **Description**
        ------------------------------      -----------------------------------------------------
        all_layers                          Optional bool. If True, all layers are cached whether
                                            they contain diagram features or not.
                                            If False, only the layers that contain diagram features
                                            are cached.
        ==============================      =====================================================

        :return:
            An array of JSON object layers with their SQL query.
            Each JSON object layer item in the array provides the following information:
            * "id"—The ID of the layer
            * "source"—The layer source internal information:
            * "type"—"workspaceLayer"
            * "workspaceID"—"Diagram"
            * "layerID"—The internal layer ID
            * "definitionExpression"-The filtering expression based on the diagram GUID

        """
        url = "%s/dynamicLayers" % self._url
        params = {
            "f": "json",
            "gdbVersion": self._version_name,
            "sessionId": self._version_guid,
            "allLayers": all_layers,
        }
        return self._con.post(url, params)

    # ----------------------------------------------------------------------
    def layer_definitions(self, all_layers: bool = False, moment: Optional[int] = None):
        """
        The Layers Definitions resource details all the layer and labeling
        properties that define each sublayer under the diagram layer.

        ==============================      =====================================================
        **Parameter**                       **Description**
        ------------------------------      -----------------------------------------------------
        all_layers                          Optional bool. If True, all layers are cached whether
                                            they contain diagram features or not.
                                            If False, only the layers that contain diagram features
                                            are cached.
        ------------------------------      -----------------------------------------------------
        moment                              Optional int. The session moment.
        ==============================      =====================================================
        """
        url = "%s/layerDefinitions" % self._url
        params = {
            "f": "json",
            "gdbVersion": self._version_name,
            "sessionId": self._version_guid,
            "allLayers": all_layers,
            "moment": moment,
        }
        return self._con.post(url, params)

    # ----------------------------------------------------------------------
    def diagram_map(self):
        """
        The Diagram Map resource mimics a map service resource. It is returned as a Map service by the REST API.

        .. note::
            This map REST endpoint is the one you can add as content to your web maps.
        """
        url = "%s/map" % self._url
        params = {
            "f": "json",
            "gdbVersion": self._version_name,
            "sessionId": self._version_guid,
        }
        return self._con.post(url, params)

    # ----------------------------------------------------------------------
    def export_diagram_map(
        self,
        size: str,
        dpi: Optional[str] = None,
        bbox: Optional[str] = None,
        format: str = "PNGS",
        transparent: bool = False,
        map_scale: Optional[str] = None,
        rotation: Optional[int] = 0,
        bbox_sr: Optional[str] = None,
        image_sr: Optional[str] = None,
        moment: Optional[int] = None,
    ):
        """
        The Export operation is performed on a Diagram Map resource.
        The result of this operation is a map image that represents the diagram specified in the URL.

        ==============================      =====================================================
        **Parameter**                       **Description**
        ------------------------------      -----------------------------------------------------
        size                                Required string. The size ("width, height") of the
                                            exported image in pixels.

                                            Example: size="600, 600"
        ------------------------------      -----------------------------------------------------
        dpi                                 Optional string. The device resolution of the exported
                                            image (dots per inch). If the dpi is not specified,
                                            an image with a default of 96 will be exported.

                                            Example: dpi="600, 600"
        ------------------------------      -----------------------------------------------------
        bbox                                Optional string. The extent(bounding box) of the exported
                                            image. Unless the bbox_sr parameter has been specified,
                                            the bbox assumed to be in the spatial reference of
                                            the map.

                                            The bbox coordinates should always use a period as a
                                            decimal separator, even in countries where traditionally
                                            a comma is used.

                                            Example: bbox="-104, 35.6, -94.32, 41"
        ------------------------------      -----------------------------------------------------
        format                              Optional string. The format of the exported image.

                                            `Values: "PNG32" | "PNG24" | "PNG" | "JPG" | "DIB" |
                                            "TIFF" | "EMF" | "PS" | "PDF" | "GIF" | "SVG" | "SVGZ" | "BMP"`
        ------------------------------      -----------------------------------------------------
        transparent                         Optional bool.  If true, the image will be exported
                                            with the background color of the map set as its
                                            transparent color. The default is false.
                                            Only the "PNG" and "GIF" formats support transparency.

                                            Internet Explorer 6 does not display transparency
                                            correctly for png24 image formats.
        ------------------------------      -----------------------------------------------------
        map_scale                           Optional int. Use this parameter to export a map image
                                            at a specific scale, with the map centered around the
                                            center of the specified bbox.
        ------------------------------      -----------------------------------------------------
        rotation                            Optional int. Use this parameter to export a map image
                                            rotated at a specific angle, with the map centered
                                            around the center of the specified bounding box (bbox).
                                            It could be a positive or negative number.
        ------------------------------      -----------------------------------------------------
        bbox_sr                             Optional string. The well-known ID that is the spatial
                                            reference of the bbox.

                                            If none is specified, the spatial reference of the
                                            map is used.
        ------------------------------      -----------------------------------------------------
        image_sr                            Optional string. The well-known ID that is the spatial
                                            reference of the exported image.

                                            If none is specified, the spatial reference of the
                                            map is used.
        ------------------------------      -----------------------------------------------------
        moment                              Optional int. The session moment.
        ==============================      =====================================================
        """
        url = "%s/map/export" % self._url
        params = {
            "f": "json",
            "gdbVersion": self._version_name,
            "moment": moment,
            "size": size,
            "dpi": dpi,
            "imageSR": image_sr,
            "bboxSR": bbox_sr,
            "format": format,
            "transparent": transparent,
            "mapScale": map_scale,
            "rotation": rotation,
            "bbox": bbox,
        }
        return self._con.post(url, params)

    # ----------------------------------------------------------------------
    def identify_diagram_map(
        self,
        geometry: Union[str, dict],
        tolerance: int,
        image_display: str,
        map_extent: str,
        geometry_precision: Optional[int] = None,
        return_field_name: bool = False,
        return_unformatted_values: bool = False,
        return_z: bool = False,
        return_m: bool = False,
        return_geometry: bool = True,
        max_allowable_offset: Optional[int] = None,
        geometry_type: str = "esriGeometryPoint",
        sr: Optional[str] = None,
        moment: Optional[int] = None,
    ):
        """
        The Identify operation is performed on a Diagram Map resource.
        The result of this operation is an identify results resource.

        ==============================      =====================================================
        **Parameter**                       **Description**
        ------------------------------      -----------------------------------------------------
        geometry                            Required string or dictionary.
                                            The geometry to identify on. The type of the geometry is specified by the geometryType parameter. The structure of the geometries is the same as the structure of the JSON geometry objects returned by the ArcGIS REST API. In addition to the JSON structures, for points and envelopes, you can specify the geometries with a simpler comma-separated syntax.

                                            Syntax:

                                            * JSON structures: geometry=<geometryType>&geometry={ geometry}
                                            * Point simple syntax: geometry=esriGeometryPoint&geometry=<x>,<y>
                                            * Envelope simple syntax: geometry=esriGeometryEnvelope&geometry=<xmin>,<ymin>,<xmax>,<ymax>


                                            Example:

                                                * JSON structures: geometry=esriGeometryPoint&geometry={x: -104, y: 35.6}
                                                * Point simple syntax:geometry=esriGeometryPoint&geometry=-104,35.6
                                                * Envelope simple syntax: geometry=esriGeometryEnvelope&geometry=-104,35.6,-94.32,41
        ------------------------------      -----------------------------------------------------
        tolerance                           Required int. The distance in screen pixels from the
                                            specified geometry within which the identify should be
                                            performed.
        ------------------------------      -----------------------------------------------------
        image_display                       Required string. The screen image display parameters
                                            (width, height, and DPI) of the map being currently viewed.

                                            The mapExtent and the imageDisplay parameters are
                                            used by the server to determine the layers visible
                                            in the current extent. They are also used to calculate
                                            the distance on the map to search based on the tolerance
                                            in screen pixels.

                                            Syntax: image_display="<width>,<heigth>,<dpi>"
        ------------------------------      -----------------------------------------------------
        map_extent                          Required string. The extent or bbox fo the map currently
                                            being viewed. Unless the sr parameter has been specified,
                                            the map_extent is assumed to be in the spatial reference
                                            of the map.

                                            The `map_extent` and `image_display` parameters are
                                            used by the server to determine the layers visible in
                                            the current extent. They are also used to calculate
                                            the distance on the map to seach based on the `tolerance`
                                            in screen pixels.

                                            Syntax: map_extent="<xmin>, <ymin>, <xmax>, <ymax>"
        ------------------------------      -----------------------------------------------------
        geometry_precision                  Optional int. Specify the number of decimal places in
                                            the response geometries returned by the operation.
                                            Does not apply to m and z values.
        ------------------------------      -----------------------------------------------------
        return_field_name                   Optional bool. If True, field names will be returned
                                            instead of field aliases.
        ------------------------------      -----------------------------------------------------
        return_unformatted_values           Optional bool. If True, the values in the result will
                                            not be formatted. Numbers will be returned as is and
                                            dates will be returned as epoch values.
        ------------------------------      -----------------------------------------------------
        return_z                            Optional bool. If True, z value will be included in
                                            the results if the features have z-values.

                                            Only applies if `return_geometry=True`
        ------------------------------      -----------------------------------------------------
        return_m                            Optional bool. If True, m value will be included in
                                            the results if the features have m-values.

                                            Only applies if `return_geometry=True`
        ------------------------------      -----------------------------------------------------
        return_geometry                     Optional bool. If True, the result set will include
                                            the geometries associated with each result.
        ------------------------------      -----------------------------------------------------
        max_allowable_offset                Optional int. This option can be used to specify the
                                            maximum allowable offset to be used for generalizing
                                            geometries returned by the identify operation.
                                            The max_allowable_offset is in the units of the sr.
                                            If sr is not specified, max_allowable_offset is
                                            assumed to be in the unit of the spatial reference of the map.
        ------------------------------      -----------------------------------------------------
        geometry type                       Optional string. The type of geometry specified by the
                                            geometry parameter. The geometry type could be a point, line, polygon, or an envelope.
                                            The default geometry type is a point ("esriGeometryPoint").

                                            `Values: "esriGeometryPoint" | "esriGeometryMultipoint" |
                                            "esriGeometryPolyline" | "esriGeometryPolygon" | "esriGeometryEnvelope"`
        ------------------------------      -----------------------------------------------------
        sr                                  Optional string. The well-known ID fo the spatial
                                            reference of the input and output geometries as well
                                            as the map_extent.

                                            If sr is not specified, the geometry and the map_extent
                                            are assumed to be in the spatial reference of the map,
                                            and the output geometries are also in the spatial reference
                                            of the map.
        ------------------------------      -----------------------------------------------------
        moment                              Optional int. The session moment.
        ==============================      =====================================================
        """
        url = "%s/map/identify" % self._url
        params = {
            "f": "json",
            "gdbVersion": self._version_name,
            "moment": moment,
            "geometry": geometry,
            "tolerance": tolerance,
            "imageDisplay": image_display,
            "mapExtent": map_extent,
            "geometryPrecision": geometry_precision,
            "returnFieldName": return_field_name,
            "returnUnformattedValues": return_unformatted_values,
            "returnZ": return_z,
            "returnM": return_m,
            "returnGeometry": return_geometry,
            "maxAllowableOffset": max_allowable_offset,
            "geometryType": geometry_type,
            "sr": sr,
        }
        return self._con.post(url, params)



import json
from pathlib import Path
from ._codetemplate import image_classifier_prf
from ._arcgis_model import _EmptyData
from functools import partial
from .._data import _raise_fastai_import_error
import traceback
import logging

logger = logging.getLogger()

try:
    from ._arcgis_model import (
        ArcGISModel,
        _resnet_family,
        _set_ddp_multigpu,
        _isnotebook,
    )
    from ._timm_utils import filter_timm_models, timm_config
    import torch
    from fastai.vision.learner import unet_learner, cnn_config
    import numpy as np
    from fastai.layers import CrossEntropyFlat
    from .._utils.segmentation_loss_functions import FocalLoss, MixUpCallback, DiceLoss
    from ._unet_utils import (
        is_no_color,
        LabelCallback,
        predict_batch,
        show_results_multispectral,
    )
    from .._utils.common import get_multispectral_data_params_from_emd, _get_emd_path
    from .._utils.classified_tiles import per_class_metrics
    from ._psp_utils import accuracy
    from ._deeplab_utils import compute_miou
    from matplotlib import pyplot as plt
    from .._utils.env import is_arcgispronotebook

    HAS_FASTAI = True
except Exception as e:
    import_exception = "\n".join(
        traceback.format_exception(type(e), e, e.__traceback__)
    )

    class NnModule:
        pass

    HAS_FASTAI = False


class UnetClassifier(ArcGISModel):
    """
    Creates a Unet like classifier based on given pretrained encoder.

    =====================   ===========================================
    **Parameter**            **Description**
    ---------------------   -------------------------------------------
    data                    Required fastai Databunch. Returned data object from
                            :meth:`~arcgis.learn.prepare_data` function.
    ---------------------   -------------------------------------------
    backbone                Optional string. Backbone convolutional neural network
                            model used for feature extraction, which
                            is `resnet34` by default.
                            Supported backbones: ResNet family and specified Timm
                            models(experimental support) from :func:`~arcgis.learn.UnetClassifier.backbones`.
    ---------------------   -------------------------------------------
    pretrained_path         Optional string. Path where pre-trained model is
                            saved.
    ---------------------   -------------------------------------------
    backend                 Optional string. Controls the backend framework to be used
                            for this model, which is 'pytorch' by default.

                            valid options are 'pytorch', 'tensorflow'
    =====================   ===========================================

    **kwargs**

    =====================   ===========================================
    **Parameter**            **Description**
    ---------------------   -------------------------------------------
    class_balancing         Optional boolean. If True, it will balance the
                            cross-entropy loss inverse to the frequency
                            of pixels per class. Default: False.
    ---------------------   -------------------------------------------
    mixup                   Optional boolean. If True, it will use mixup
                            augmentation and mixup loss. Default: False
    ---------------------   -------------------------------------------
    focal_loss              Optional boolean. If True, it will use focal loss
                            Default: False
    ---------------------   -------------------------------------------
    dice_loss_fraction      Optional float.
                            Min_val=0, Max_val=1
                            If > 0 , model will use a combination of default or
                            focal(if focal=True) loss with the specified fraction
                            of dice loss.
                            E.g.
                            for dice = 0.3, loss = (1-0.3)*default loss + 0.3*dice
                            Default: 0
    ---------------------   -------------------------------------------
    dice_loss_average       Optional str.
                            micro: Micro dice coefficient will be used for loss
                            calculation.
                            macro: Macro dice coefficient will be used for loss
                            calculation.
                            A macro-average will compute the metric independently
                            for each class and then take the average (hence treating
                            all classes equally), whereas a micro-average will
                            aggregate the contributions of all classes to compute the
                            average metric. In a multi-class classification setup,
                            micro-average is preferable if you suspect there might be
                            class imbalance (i.e you may have many more examples of
                            one class than of other classes)
                            Default: 'micro'
    ---------------------   -------------------------------------------
    ignore_classes          Optional list. It will contain the list of class
                            values on which model will not incur loss.
                            Default: []
    =====================   ===========================================

    :return: :class:`~arcgis.learn.UnetClassifier` Object
    """

    def __init__(
        self,
        data,
        backbone=None,
        pretrained_path=None,
        backend="pytorch",
        *args,
        **kwargs,
    ):
        if pretrained_path is not None:
            backbone_pretrained = False
        else:
            backbone_pretrained = True

        self._backend = backend
        if self._backend == "tensorflow":
            super().__init__(data, None)
            self._intialize_tensorflow(data, backbone, pretrained_path, kwargs)
        else:
            super().__init__(data, backbone, pretrained_path=pretrained_path, **kwargs)
            data = self._data

            self._check_dataset_support(self._data)
            if not (self._check_backbone_support(getattr(self, "_backbone", backbone))):
                raise Exception(
                    f"Enter only compatible backbones from {', '.join(self.supported_backbones)}"
                )

            self._ignore_classes = kwargs.get("ignore_classes", [])
            if self._ignore_classes != [] and len(data.classes) <= 3:
                raise Exception(
                    f"`ignore_classes` parameter can only be used when the dataset has more than 2 classes."
                )

            data_classes = list(self._data.class_mapping.keys())
            if 0 not in list(data.class_mapping.values()):
                self._ignore_mapped_class = [
                    data_classes.index(k) + 1 for k in self._ignore_classes if k != 0
                ]
            else:
                self._ignore_mapped_class = [
                    data_classes.index(k) + 1 for k in self._ignore_classes
                ]
            if self._ignore_classes != []:
                if 0 not in self._ignore_mapped_class:
                    self._ignore_mapped_class.insert(0, 0)
                global accuracy
                accuracy = partial(
                    accuracy, ignore_mapped_class=self._ignore_mapped_class
                )

            self.mixup = kwargs.get("mixup", False)
            self.class_balancing = kwargs.get("class_balancing", False)
            self.focal_loss = kwargs.get("focal_loss", False)
            self.dice_loss_fraction = kwargs.get("dice_loss_fraction", False)
            self.weighted_dice = kwargs.get("weighted_dice", False)
            self.dice_loss_average = kwargs.get("dice_loss_average", "micro")

            self._code = image_classifier_prf

            backbone_cut = None
            backbone_split = None

            if hasattr(self, "_orig_backbone"):
                _backbone_meta = cnn_config(self._orig_backbone)
                backbone_cut = _backbone_meta["cut"]
                backbone_split = _backbone_meta["split"]

            if "timm" in self._backbone.__module__:
                for bckbn in ["densenet", "inception_v4", "vgg"]:
                    if bckbn in self._backbone.__name__:
                        from torch import nn
                        from fastai.vision.learner import has_pool_type

                        def bckbn_cut(m):
                            ll = list(enumerate(m.children()))
                            cut = next(i for i, o in reversed(ll) if has_pool_type(o))
                            m = nn.Sequential(*list(m.children())[:cut])
                            return m[0]

                        backbone_cut = bckbn_cut
                        backbone_split = None
                        break
                else:
                    timm_meta = timm_config(self._backbone)
                    backbone_cut = timm_meta["cut"]
                    backbone_split = timm_meta["split"]

                if (
                    "nasnet" in self._backbone.__name__
                    or "repvgg" in self._backbone.__name__
                ):
                    backbone_cut = None

            if not _isnotebook():
                _set_ddp_multigpu(self)
                if self._multigpu_training:
                    self.learn = unet_learner(
                        data,
                        arch=self._backbone,
                        pretrained=backbone_pretrained,
                        metrics=accuracy,
                        wd=1e-2,
                        bottle=True,
                        last_cross=True,
                        cut=backbone_cut,
                        split_on=backbone_split,
                    ).to_distributed(self._rank_distributed)
                    self._map_location = {
                        "cuda:%d" % 0: "cuda:%d" % self._rank_distributed
                    }
                else:
                    self.learn = unet_learner(
                        data,
                        arch=self._backbone,
                        pretrained=backbone_pretrained,
                        metrics=accuracy,
                        wd=1e-2,
                        bottle=True,
                        last_cross=True,
                        cut=backbone_cut,
                        split_on=backbone_split,
                    )
            else:
                self.learn = unet_learner(
                    data,
                    arch=self._backbone,
                    pretrained=backbone_pretrained,
                    metrics=accuracy,
                    wd=1e-2,
                    bottle=True,
                    last_cross=True,
                    cut=backbone_cut,
                    split_on=backbone_split,
                )

            class_weight = None
            if self.class_balancing:
                if data.class_weight is not None:
                    # Handle condition when nodata is already at pixel value 0 in data
                    if (data.c - 1) == data.class_weight.shape[0]:
                        class_weight = (
                            torch.tensor(
                                [data.class_weight.mean()] + data.class_weight.tolist()
                            )
                            .float()
                            .to(self._device)
                        )
                    else:
                        class_weight = (
                            torch.tensor(data.class_weight).float().to(self._device)
                        )
                else:
                    if getattr(data, "overflow_encountered", False):
                        logger.warning(
                            "Overflow Encountered. Ignoring `class_balancing` parameter."
                        )
                        class_weight = [1] * len(data.classes)
                    else:
                        logger.warning(
                            "Could not find 'NumPixelsPerClass' in 'esri_accumulated_stats.json'. Ignoring `class_balancing` parameter."
                        )

            if self._ignore_classes != []:
                if not self.class_balancing:
                    class_weight = torch.tensor([1] * data.c).float().to(self._device)
                class_weight[self._ignore_mapped_class] = 0.0

            self._final_class_weight = class_weight
            self.learn.loss_func = CrossEntropyFlat(class_weight, axis=1)

            if self.focal_loss:
                self.learn.loss_func = FocalLoss(self.learn.loss_func)
            if self.mixup:
                self.learn.callbacks.append(MixUpCallback(self.learn))

            if self.dice_loss_fraction:
                self.learn.loss_func = DiceLoss(
                    self.learn.loss_func,
                    self.dice_loss_fraction,
                    weighted_dice=self.weighted_dice,
                    dice_average=self.dice_loss_average,
                )

            self._arcgis_init_callback()  # make first conv weights learnable
            self.learn.callbacks.append(
                LabelCallback(self.learn)
            )  # appending label callback

            self.learn.model = self.learn.model.to(self._device)
            # _set_multigpu_callback(self) # MultiGPU doesn't work for U-Net. (Fastai-Forums)
            if pretrained_path is not None:
                self.load(pretrained_path)

    def __str__(self):
        return self.__repr__()

    def __repr__(self):
        return "<%s>" % (type(self).__name__)

    @staticmethod
    def _available_metrics():
        return ["valid_loss", "accuracy"]

    @property
    def supported_backbones(self):
        """Supported list of backbones for this model."""
        return UnetClassifier._supported_backbones()

    @staticmethod
    def backbones():
        """Supported list of backbones for this model."""
        return UnetClassifier._supported_backbones()

    @staticmethod
    def _supported_backbones():
        timm_models = filter_timm_models(
            [
                "*dpn*",
                "*hrnet*",
                "nasnetalarge",
                "pnasnet5large",
                "*repvgg*",
                "*selecsls*",
                "*tresnet*",
            ]
        )
        timm_backbones = list(map(lambda m: "timm:" + m, timm_models))
        return [*_resnet_family] + timm_backbones

    @property
    def supported_datasets(self):
        """Supported dataset types for this model."""
        return UnetClassifier._supported_datasets()

    @staticmethod
    def _supported_datasets():
        return ["Classified_Tiles"]

    @classmethod
    def from_model(cls, emd_path, data=None):
        """
        Creates a Unet like classifier from an Esri Model Definition (EMD) file.

        =====================   ===========================================
        **Parameter**            **Description**
        ---------------------   -------------------------------------------
        emd_path                Required string. Path to Deep Learning Package
                                (DLPK) or Esri Model Definition(EMD) file.
        ---------------------   -------------------------------------------
        data                    Required fastai Databunch or None. Returned data
                                object from :meth:`~arcgis.learn.prepare_data` function or None for
                                inferencing.
        =====================   ===========================================

        :return: :class:`~arcgis.learn.UnetClassifier` Object
        """
        return cls.from_emd(data, emd_path)

    @classmethod
    def from_emd(cls, data, emd_path):
        """
        Creates a Unet like classifier from an Esri Model Definition (EMD) file.

        =====================   ===========================================
        **Parameter**            **Description**
        ---------------------   -------------------------------------------
        data                    Required fastai Databunch or None. Returned data
                                object from :meth:`~arcgis.learn.prepare_data` function or None for
                                inferencing.
        ---------------------   -------------------------------------------
        emd_path                Required string. Path to Esri Model Definition
                                file.
        =====================   ===========================================

        :return: :class:`~arcgis.learn.UnetClassifier` Object
        """
        if not HAS_FASTAI:
            _raise_fastai_import_error(import_exception=import_exception)

        emd_path = _get_emd_path(emd_path)
        with open(emd_path) as f:
            emd = json.load(f)

        model_file = Path(emd["ModelFile"])

        if not model_file.is_absolute():
            model_file = emd_path.parent / model_file

        model_params = emd["ModelParameters"]

        try:
            class_mapping = {i["Value"]: i["Name"] for i in emd["Classes"]}
            color_mapping = {i["Value"]: i["Color"] for i in emd["Classes"]}
        except KeyError:
            class_mapping = {i["ClassValue"]: i["ClassName"] for i in emd["Classes"]}
            color_mapping = {i["ClassValue"]: i["Color"] for i in emd["Classes"]}

        resize_to = emd.get("resize_to")

        if data is None:
            data = _EmptyData(
                path=emd_path.parent.parent,
                loss_func=None,
                c=len(class_mapping) + 1,
                chip_size=emd["ImageHeight"],
            )
            data.class_mapping = class_mapping
            data.color_mapping = color_mapping
            data = get_multispectral_data_params_from_emd(data, emd)

            data.emd_path = emd_path
            data.emd = emd
            data._is_empty = True

        data.resize_to = resize_to

        return cls(data, **model_params, pretrained_path=str(model_file))

    @property
    def _model_metrics(self):
        return {"accuracy": "{0:1.4e}".format(self._get_model_metrics())}

    def _get_emd_params(self, save_inference_file):
        import random

        _emd_template = {}
        _emd_template["Framework"] = "arcgis.learn.models._inferencing"
        _emd_template["ModelConfiguration"] = "_unet"
        _emd_template["ModelType"] = "ImageClassification"
        if save_inference_file:
            _emd_template["InferenceFunction"] = "ArcGISImageClassifier.py"
        else:
            _emd_template[
                "InferenceFunction"
            ] = "[Functions]System\\DeepLearning\\ArcGISLearn\\ArcGISImageClassifier.py"
        _emd_template["ExtractBands"] = [0, 1, 2]
        _emd_template["ignore_mapped_class"] = self._ignore_mapped_class
        _emd_template["SupportsVariableTileSize"] = True

        _emd_template["Classes"] = []
        class_data = {}
        for i, class_name in enumerate(
            self._data.classes[1:]
        ):  # 0th index is background
            inverse_class_mapping = {v: k for k, v in self._data.class_mapping.items()}
            class_data["Value"] = inverse_class_mapping[class_name]
            class_data["Name"] = class_name
            color = (
                [random.choice(range(256)) for i in range(3)]
                if is_no_color(self._data.color_mapping)
                else self._data.color_mapping[inverse_class_mapping[class_name]]
            )
            class_data["Color"] = color
            _emd_template["Classes"].append(class_data.copy())

        return _emd_template

    def _predict_batch(self, imagetensor_batch):
        return predict_batch(self, imagetensor_batch)

    def _show_results_multispectral(
        self, rows=5, alpha=0.7, **kwargs
    ):  # parameters adjusted in kwargs
        """
        Shows the ground truth and predictions of model side by side.

        **kwargs**

        =====================   ===========================================
        **Parameter**            **Description**
        ---------------------   -------------------------------------------
        rows                    Number of rows of data to be displayed, if
                                batch size is smaller, then the rows will
                                display the value provided for batch size.
        ---------------------   -------------------------------------------
        alpha                   Optional Float. Opacity parameter for label
                                overlay on image. Float [0.0 - 1.0]
                                Default: 0.7
        =====================   ===========================================

        """
        return_fig = kwargs.get("return_fig", False)
        ret_val = show_results_multispectral(self, nrows=rows, alpha=alpha, **kwargs)
        if return_fig:
            fig, ax = ret_val
            return fig

    def show_results(self, rows=5, **kwargs):
        """
        Displays the results of a trained model on a part of the validation set.

        =====================   ===========================================
        **Parameter**            **Description**
        ---------------------   -------------------------------------------
        rows                    Optional int. Number of rows of results
                                to be displayed.
        =====================   ===========================================

        """
        self._check_requisites()
        self.learn.callbacks = [
            x for x in self.learn.callbacks if not isinstance(x, LabelCallback)
        ]
        if rows > len(self._data.valid_ds):
            rows = len(self._data.valid_ds)
        self.learn.show_results(
            rows=rows, ignore_mapped_class=self._ignore_mapped_class, **kwargs
        )
        if is_arcgispronotebook():
            plt.show()

    def accuracy(self):
        """Computes per pixel accuracy on validation set."""
        try:
            return self.learn.validate()[1].tolist()
        except Exception as e:
            accuracy = self._data.emd.get("accuracy")
            if accuracy:
                return accuracy
            else:
                logger.error("Metric not found in the loaded model")

    def _get_model_metrics(self, **kwargs):
        checkpoint = getattr(self, "_is_checkpointed", False)
        if not hasattr(self.learn, "recorder"):
            return 0.0

        try:
            model_accuracy = self.learn.recorder.metrics[-1][0]
            if checkpoint:
                val_losses = self.learn.recorder.val_losses
                model_accuracy = self.learn.recorder.metrics[
                    val_losses.index(min(val_losses))
                ][0]
        except:
            model_accuracy = 0.0

        return float(model_accuracy)

    def mIOU(self, mean=False, show_progress=True):
        """
        Computes mean IOU on the validation set for each class.

        =====================   ===========================================
        **Parameter**            **Description**
        ---------------------   -------------------------------------------
        mean                    Optional bool. If False returns class-wise
                                mean IOU, otherwise returns mean iou of all
                                classes combined.
        ---------------------   -------------------------------------------
        show_progress           Optional bool. Displays the progress bar if
                                True.
        =====================   ===========================================

        :return: `dict` if mean is False otherwise `float`
        """
        self._check_requisites()
        num_classes = torch.arange(self._data.c)
        miou = compute_miou(
            self,
            self._data.valid_dl,
            mean,
            num_classes,
            show_progress,
            self._ignore_mapped_class,
        )
        if mean:
            miou = [
                miou[i] for i in range(len(miou)) if i not in self._ignore_mapped_class
            ]
            return np.mean(miou)
        if self._ignore_mapped_class == []:
            return dict(zip(["0"] + self._data.classes[1:], miou))
        else:
            class_values = [0] + list(self._data.class_mapping.keys())
            return {
                class_values[i]: miou[i]
                for i in range(len(miou))
                if i not in self._ignore_mapped_class
            }

    ## Tensorflow specific functions start ##
    def _intialize_tensorflow(self, data, backbone, pretrained_path, kwargs):
        self._check_tf()
        self._ignore_mapped_class = []

        import tensorflow as tf
        from .._utils.common import get_color_array
        from .._utils.common_tf import handle_backbone_parameter, get_input_shape
        from .._model_archs.unet_tf import get_unet_tf_model
        from tensorflow.keras.losses import (
            SparseCategoricalCrossentropy,
            BinaryCrossentropy,
        )
        from .._utils.fastai_tf_fit import TfLearner, defaults
        from tensorflow.keras.models import Model
        from tensorflow.keras.optimizers import Adam
        from .._utils.common import kwarg_fill_none

        if data._is_multispectral:
            raise Exception(
                'Multispectral data is not supported with backend="tensorflow"'
            )

        # Intialize Tensorflow
        self._init_tensorflow(data, backbone)

        # Loss Function
        # self._loss_function_tf_ = BinaryCrossentropy(from_logits=True)
        self._loss_function_tf_ = SparseCategoricalCrossentropy(
            from_logits=True, reduction="auto"
        )

        self._mobile_optimized = kwarg_fill_none(
            kwargs, "mobile_optimized", self._backbone_mobile_optimized
        )

        # Create Unet Model
        model = get_unet_tf_model(
            self._backbone_initalized, data, mobile_optimized=self._mobile_optimized
        )

        self.learn = TfLearner(
            data,
            model,
            opt_func=Adam,
            loss_func=self._loss_function_tf,
            true_wd=True,
            bn_wd=True,
            wd=defaults.wd,
            train_bn=True,
        )

        self.learn.unfreeze()
        self.learn.freeze_to(len(self._backbone_initalized.layers))

        self.show_results = self._show_results_multispectral

        self._code = image_classifier_prf

    def _loss_function_tf(self, target, predictions):
        import tensorflow as tf

        # print(target.shape, predictions.shape)
        # print(target.dtype, predictions.dtype)
        # print(tf.unique(tf.reshape(target, [-1]))[0])
        # print('\n', tf.unique(tf.reshape(predictions, [-1]))[0])
        # print(tf.unique(tf.reshape(target, [-1])).numpy(), tf.unique(tf.reshape(predictions, [-1])))
        target = tf.squeeze(target, axis=1)

        # from .._utils.pixel_classification import segmentation_mask_to_one_hot
        # from .._utils.fastai_tf_fit import _pytorch_to_tf
        # target = _pytorch_to_tf(segmentation_mask_to_one_hot(target.cpu().numpy(), self._data.c).permute(0, 2, 3, 1))

        return self._loss_function_tf_(target, predictions)

    ## Tensorflow specific functions end ##

    def per_class_metrics(self, ignore_classes=[]):
        """
        Computer per class precision, recall and f1-score on validation set.

        =====================   ===========================================
        **Parameter**            **Description**
        ---------------------   -------------------------------------------
        self                    segmentation model object -> [PSPNetClassifier | UnetClassifier | DeepLab]
        ---------------------   -------------------------------------------
        ignore_classes          Optional list. It will contain the list of class
                                values on which model will not incur loss.
                                Default: []
        =====================   ===========================================

        Returns per class precision, recall and f1 scores
        """
        ignore_classes = np.unique(self._ignore_classes + ignore_classes).tolist()
        try:
            self._check_requisites()
            ## Calling imported function `per_class_metrics`
            return per_class_metrics(self, ignore_classes)
        except:
            import pandas as pd

            if "per_class_metrics" in self._data.emd.keys():
                return pd.read_json(self._data.emd["per_class_metrics"])
            else:
                logger.error("Metric not found in the loaded model")



"""
The ``Filters`` modules contain functions to filter query results by a spatial relationship with another
:class:`~arcgis.geometry.Geometry` object. The ``Filters`` module is used when querying feature layers and imagery
layers.
"""
from typing import Union
from arcgis.geometry._types import Geometry, SpatialReference


# esriSpatialRelIntersects | esriSpatialRelContains | esriSpatialRelCrosses | esriSpatialRelEnvelopeIntersects | \
# esriSpatialRelIndexIntersects | esriSpatialRelOverlaps | esriSpatialRelTouches | esriSpatialRelWithin


def _filter(geometry, sr, rel):
    if not isinstance(geometry, Geometry):
        geometry = Geometry(geometry)

    gt = {
        "point": "Point",
        "multipoint": "Multipoint",
        "polygon": "Polygon",
        "polyline": "Polyline",
        "envelope": "Envelope",
    }

    filter = {
        "geometry": geometry,
        "geometryType": "esriGeometry" + gt[str(geometry.type).lower()],
        "spatialRel": rel,
    }

    if sr is None:
        if "spatialReference" in geometry:
            sr = geometry["spatialReference"]

    if sr is not None:
        filter["inSR"] = sr

    return filter


def intersects(geometry: Geometry, sr: Union[SpatialReference, dict, None] = None):
    """
    The ``intersects`` method filters results whose geometry intersects with the specified
    :class:`~arcgis.geometry.Geometry` object.

    ================  ===============================================================================
    **Keys**          **Description**
    ----------------  -------------------------------------------------------------------------------
    geometry          A single :class:`~arcgis.geometry.Geometry` of any type. The structure of
                      geometry is the same as the structure of the JSON geometry
                      objects returned by the ArcGIS REST API. The use of simple
                      syntax is not supported.
    ----------------  -------------------------------------------------------------------------------
    spatial_ref       A :class:`~arcgis.geometry.SpatialReference` of the input geometries Well-Known ID or JSON object
    ================  ===============================================================================

    :return: A dictionary

    .. code-block:: python

        USAGE EXAMPLE: Select the gas lines that intersect a specific
        freeway feature, United States Interstate 15.

        from arcgis.geometry import Geometry
        from arcgis.geometry.filters import intersects

        # get a feature layer item from the gis
        >>> flyr_item = gis.content.search("Freeways*", "Feature Layer")[0]
        >>> freeway_lyr = flyr_item.layers[0]

        # assign a variable to the spatial reference
        >>> freeway_sr = freeway_lyr.properties.extent["spatialReference"]

        # select a filter feature to construct its geometry
        >>> rte15_fset = freeway_lyr.query(where="ROUTE_NUM = 'I15'")
        >>> rte15_geom_dict = rte15_fset.features[0].geometry
        >>> rte15_geom_dict['spatialReference'] = freeway_sr
        >>> rte15_geom = Geometry(rte15_geom_dict)

        # construct a geometry filter using the filter geometry
        >>> flyr_filter = intersects(rte15_geom, sr=freeway_sr)

        # query a feature layer for features that meet filter criteria
        >>> gas_lines_I15 = gas_line_lyr.query(geometry_filter=flyr_filter)
    """

    return _filter(geometry, sr, "esriSpatialRelIntersects")


def contains(geometry: Geometry, sr: Union[SpatialReference, dict, None] = None):
    """
    The ``contains`` method returns a feature if its shape is wholly contained within the search
    :class:`~arcgis.geometry.Geometry` object.

    .. note::
        Valid for all shape type combinations.

    ================  ===============================================================================
    **Keys**          **Description**
    ----------------  -------------------------------------------------------------------------------
    geometry          A single :class:`~arcgis.geometry.Geometry` of any type. The structure of
                      geometry is the same as the structure of the JSON geometry
                      objects returned by the ArcGIS REST API. The use of simple
                      syntax is not supported.
    ----------------  -------------------------------------------------------------------------------
    spatial_ref       A :class:`~arcgis.geometry.SpatialReference` of the input geometries Well-Known ID or JSON object
    ================  ===============================================================================

    :return: A dictionary
    """
    return _filter(geometry, sr, "esriSpatialRelContains")


def crosses(geometry: Geometry, sr: Union[SpatialReference, dict, None] = None):
    """
    The ``crosses`` method retrieves a feature if the intersection of the interiors of the two shapes is not empty
    **and** has a lower dimension than the maximum dimension of the two shapes. Two lines that share an endpoint in
    common do not cross.

    .. note::
        Valid for Line/Line, Line/Area, :class:`~arcgis.geometry.MultiPoint` /Area, and
        :class:`~arcgis.geometry.MultiPoint` /Line shape type combinations.

    ================  ===============================================================================
    **Keys**          **Description**
    ----------------  -------------------------------------------------------------------------------
    geometry          A single :class:`~arcgis.geometry.Geometry` of any type. The structure of
                      geometry is the same as the structure of the JSON geometry
                      objects returned by the ArcGIS REST API. The use of simple
                      syntax is not supported.
    ----------------  -------------------------------------------------------------------------------
    spatial_ref       A :class:`~arcgis.geometry.SpatialReference` of the input geometries Well-Known ID or JSON object
    ================  ===============================================================================

    :return: A dictionary
    """
    return _filter(geometry, sr, "esriSpatialRelCrosses")


def envelope_intersects(
    geometry: Geometry, sr: Union[SpatialReference, dict, None] = None
):
    """
    The ``envelope_intersects`` retrieves features if the :class:`~arcgis.geometry.Envelope` of the two shapes
    intersects.

    ================  ===============================================================================
    **Keys**          **Description**
    ----------------  -------------------------------------------------------------------------------
    geometry          A single :class:`~arcgis.geometry.Geometry` of any type. The structure of
                      geometry is the same as the structure of the JSON geometry
                      objects returned by the ArcGIS REST API. The use of simple
                      syntax is not supported.
    ----------------  -------------------------------------------------------------------------------
    spatial_ref       A :class:`~arcgis.geometry.SpatialReference` of the input geometries Well-Known ID or JSON object
    ================  ===============================================================================

    :return: A dictionary
    """
    return _filter(geometry, sr, "esriSpatialRelEnvelopeIntersects")


def index_intersects(
    geometry: Geometry, sr: Union[SpatialReference, dict, None] = None
):
    """
    The ``index_intersects`` method retrieves a feature if the :class:`~arcgis.geometry.Envelope` of the query
    :class:`~arcgis.geometry.Geometry` intersects the index entry for the target geometry.

    ================  ===============================================================================
    **Keys**          **Description**
    ----------------  -------------------------------------------------------------------------------
    geometry          A single :class:`~arcgis.geometry.Geometry` of any type. The structure of
                      geometry is the same as the structure of the JSON geometry
                      objects returned by the ArcGIS REST API. The use of simple
                      syntax is not supported.
    ----------------  -------------------------------------------------------------------------------
    spatial_ref       A :class:`~arcgis.geometry.SpatialReference` of the input geometries Well-Known ID or JSON object
    ================  ===============================================================================

    :return: A dictionary
    """
    return _filter(geometry, sr, "esriSpatialRelIndexIntersects")


def overlaps(geometry: Geometry, sr: Union[SpatialReference, dict, None] = None):
    """
    The ``overlaps`` method retrieves a feature if the intersection of the two shapes results in an object of the same
    dimension, but different from both of the shapes.

    .. note::
        This applies to Area/Area, Line/Line, and Multi-point/Multi-point shape type combinations.

    ================  ===============================================================================
    **Keys**          **Description**
    ----------------  -------------------------------------------------------------------------------
    geometry          A single :class:`~arcgis.geometry.Geometry` of any type. The structure of
                      geometry is the same as the structure of the JSON geometry
                      objects returned by the ArcGIS REST API. The use of simple
                      syntax is not supported.
    ----------------  -------------------------------------------------------------------------------
    spatial_ref       A :class:`~arcgis.geometry.SpatialReference` of the input geometries Well-Known ID or JSON object
    ================  ===============================================================================

    :return: A dictionary
    """
    return _filter(geometry, sr, "esriSpatialRelOverlaps")


def touches(geometry: Geometry, sr: Union[SpatialReference, dict, None] = None):
    """
    The ``touches`` method retrieves a feature if the two shapes share a common boundary. However, the intersection of
    the interiors of the two shapes must be empty.

    .. note::
        In the Point/Line case, the point may touch an endpoint only of the line. Applies to all
        combinations except Point/Point.

    ================  ===============================================================================
    **Keys**          **Description**
    ----------------  -------------------------------------------------------------------------------
    geometry          A single :class:`~arcgis.geometry.Geometry` of any type. The structure of
                      geometry is the same as the structure of the JSON geometry
                      objects returned by the ArcGIS REST API. The use of simple
                      syntax is not supported.
    ----------------  -------------------------------------------------------------------------------
    spatial_ref       A :class:`~arcgis.geometry.SpatialReference` of the input geometries Well-Known ID or JSON object
    ================  ===============================================================================

    :return: A dictionary
    """
    return _filter(geometry, sr, "esriSpatialRelTouches")


def within(geometry: Geometry, sr: Union[SpatialReference, dict, None] = None):
    """
     The ``within`` method retrieves a feature if its shape wholly contains the search
     :class:`~arcgis.geometry.Geometry`.

     .. note::
        Valid for all shape type combinations.

    ================  ===============================================================================
    **Keys**          **Description**
    ----------------  -------------------------------------------------------------------------------
    geometry          A single :class:`~arcgis.geometry.Geometry` of any type. The structure of
                      geometry is the same as the structure of the JSON geometry
                      objects returned by the ArcGIS REST API. The use of simple
                      syntax is not supported.
    ----------------  -------------------------------------------------------------------------------
    spatial_ref       A :class:`~arcgis.geometry.SpatialReference` of the input geometries Well-Known ID or JSON object
    ================  ===============================================================================

    :return: A dictionary

    """
    return _filter(geometry, sr, "esriSpatialRelWithin")
